import sys, optparse, os.path, webbrowser
import pdb, traceback
import snakes.plugins
from snakes.utils.abcd.build import Builder
from snakes.lang.abcd.parser import parse
from snakes.lang.pgen import ParseError
from snakes.utils.abcd import CompilationError, DeclarationError
from snakes.utils.abcd.simul import Simulator, ABCDSimulator
from snakes.utils.abcd.checker import Checker
from snakes.utils.abcd.html import build as html
from snakes.utils.simul.html import json

##
## error messages
##

options = None

ERR_ARG = 1
ERR_OPT = 2
ERR_IO = 3
ERR_PARSE = 4
ERR_PLUGIN = 5
ERR_COMPILE = 6
ERR_OUTPUT = 7
ERR_BUG = 255

def log (message) :
    sys.stdout.write("abcd: %s\n" % message.strip())
    sys.stdout.flush()

def err (message) :
    sys.stderr.write("abcd: %s\n" % message.strip())
    sys.stderr.flush()

def die (code, message=None) :
    global options
    if message :
        err(message)
    if options.debug :
        pdb.post_mortem(sys.exc_info()[2])
    else :
        sys.exit(code)

def bug () :
    global options
    sys.stderr.write("""
    ********************************************************************
    *** An unexpected error ocurred. Please report this bug to       ***
    *** <franck.pommereau@gmail.com>, together with the execution    ***
    *** trace below and, if possible, a stripped-down version of the ***
    *** ABCD source code that caused this bug. Thank you for your    ***
    *** help in improving SNAKES!                                    ***
    ********************************************************************

""")
    traceback.print_exc()
    if options.debug :
        pdb.post_mortem(sys.exc_info()[2])
    else :
        sys.exit(ERR_BUG)

##
## options parsing
##

gv_engines = ("dot", "neato", "twopi", "circo", "fdp")

opt = optparse.OptionParser(prog="abcd",
                            usage="%prog [OPTION]... FILE")
opt.add_option("-l", "--load",
               dest="plugins", action="append", default=[],
               help="load plugin (this option can be repeated)",
               metavar="PLUGIN")
opt.add_option("-p", "--pnml",
               dest="pnml", action="store", default=None,
               help="save net as PNML",
               metavar="OUTFILE")
for engine in gv_engines :
    opt.add_option("-" + engine[0], "--" + engine,
                   dest="gv" + engine, action="store", default=None,
                   help="draw net using '%s' (from GraphViz)" % engine,
                   metavar="OUTFILE")
opt.add_option("-a", "--all-names",
               dest="allnames", action="store_true", default=False,
               help="draw control-flow places names (default: hide)")
opt.add_option("--debug",
               dest="debug", action="store_true", default=False,
               help="launch debugger on compiler error (default: no)")
opt.add_option("--progress",
               dest="progress", action="store_true", default=False,
               help="show progression during long operations (default: no)")
opt.add_option("-s", "--simul",
               dest="simul", action="store_true", default=False,
               help="launch interactive code simulator")
opt.add_option("--headless",
               dest="headless", action="store", default=None,
               help="headless code simulator, with saved parameters",
               metavar="JSONFILE")
opt.add_option("--port",
               dest="port", action="store", default=8000, type=int,
               help="port on which the simulator server runs",
               metavar="PORT")
opt.add_option("-H", "--html",
               dest="html", action="store", default=None,
               help="save net as HTML",
               metavar="OUTFILE")
opt.add_option("--check",
               dest="check", action="store_true", default=False,
               help="check assertions")

def getopts (args) :
    global options, abcd, tmp
    (options, args) = opt.parse_args(args)
    plugins = []
    for p in options.plugins :
        plugins.extend(t.strip() for t in p.split(","))
    if "ops" not in options.plugins :
        plugins.append("ops")
    if "labels" not in plugins :
        plugins.append("labels")
    for engine in gv_engines :
        gvopt = getattr(options, "gv%s" % engine)
        if gvopt and "gv" not in plugins :
            plugins.append("gv")
            break
    if (options.html or options.simul) and "gv" not in plugins :
        plugins.append("gv")
    options.plugins = plugins
    if len(args) < 1 :
        err("no input file provided")
        opt.print_help()
        die(ERR_ARG)
    elif len(args) > 1 :
        err("more than one input file provided")
        opt.print_help()
        die(ERR_ARG)
    abcd = args[0]
    if options.pnml == abcd :
        err("input file also used as output (--pnml)")
        opt.print_help()
        die(ERR_ARG)
    if options.html == abcd :
        err("input file also used as output (--html)")
        opt.print_help()
        die(ERR_ARG)
    for engine in gv_engines :
        if getattr(options, "gv%s" % engine) == abcd :
            err("input file also used as output (--%s)" % engine)
            opt.print_help()
            die(ERR_ARG)

##
## drawing nets
##

def place_attr (place, attr) :
    # fix color
    if place.status == snk.entry :
        attr["fillcolor"] = "green"
    elif place.status == snk.internal :
        pass
    elif place.status == snk.exit :
        attr["fillcolor"] = "orange"
    else :
        attr["fillcolor"] = "lightblue"
    # fix shape
    if (not options.allnames
        and place.status in (snk.entry, snk.internal, snk.exit)) :
        attr["shape"] = "circle"
    # render marking
    if place._check == snk.tBlackToken :
        count = len(place.tokens)
        if count == 0 :
            marking = " "
        elif count == 1 :
            marking = "&#8226;"
        else :
            marking = "%s&#8226;" % count
    else :
        marking = str(place.tokens)
    # node label
    if (options.allnames
        or place.status not in (snk.entry, snk.internal, snk.exit)) :
        attr["label"] = "%s\\n%s" % (place.name, marking)
    else :
        attr["label"] = "%s" % marking

def trans_attr (trans, attr) :
    pass

def arc_attr (label, attr) :
    if label == snk.Value(snk.dot) :
        del attr["label"]
    elif isinstance(label, snk.Test) :
        attr["arrowhead"] = "none"
        attr["label"] = " %s " % label._annotation
    elif isinstance(label, snk.Flush) :
        attr["arrowhead"] = "box"
        attr["label"] = " %s " % label._annotation


def draw (net, target, engine="dot") :
    try :
        return net.draw(target, engine=engine,
                        place_attr=place_attr,
                        trans_attr=trans_attr,
                        arc_attr=arc_attr)
    except :
        die(ERR_OUTPUT, str(sys.exc_info()[1]))

##
## save pnml
##

def save_pnml (net, target) :
    try :
        out = open(target, "w")
        out.write(snk.dumps(net))
        out.close()
    except :
        die(ERR_OUTPUT, str(sys.exc_info()[1]))

##
## simulator (not standalone)
##

def simulate (source, filename="<string>") :
    global options, snk
    getopts(["--simul", filename])
    node = parse(source, filename=filename)
    snk = snakes.plugins.load(options.plugins, "snakes.nets", "snk")
    build = Builder(snk)
    net = build.build(node)
    net.label(srcfile=filename, snakes=snk)
    return ABCDSimulator(node, net, draw(net, None))

##
## main
##

def main (args=sys.argv[1:], src=None) :
    global options, snk
    # get options
    try:
        if src is None :
            getopts(args)
        else :
            getopts(list(args) + ["<string>"])
    except SystemExit :
        raise
    except :
        die(ERR_OPT, str(sys.exc_info()[1]))
    # read source
    if src is not None :
        source = src
    else :
        try :
            source = open(abcd).read()
        except :
            die(ERR_IO, "could not read input file %r" % abcd)
    # parse
    try :
        node = parse(source, filename=abcd)
    except ParseError :
        die(ERR_PARSE, str(sys.exc_info()[1]))
    except :
        bug()
    # compile
    dirname = os.path.dirname(abcd)
    if dirname and dirname not in sys.path :
        sys.path.append(dirname)
    elif "." not in sys.path :
        sys.path.append(".")
    try :
        snk = snakes.plugins.load(options.plugins, "snakes.nets", "snk")
    except :
        die(ERR_PLUGIN, str(sys.exc_info()[1]))
    build = Builder(snk)
    try :
        net = build.build(node)
        net.label(srcfile=abcd, snakes=snk)
    except (CompilationError, DeclarationError) :
        die(ERR_COMPILE, str(sys.exc_info()[1]))
    except :
        bug()
    # output
    if options.pnml :
        save_pnml(net, options.pnml)
    for engine in gv_engines :
        target = getattr(options, "gv%s" % engine)
        if target :
            draw(net, target, engine)
    if options.html :
        try :
            html(abcd, node, net, draw(net, None), options.html)
        except :
            bug()
    trace, lineno = [], None
    if options.check :
        states, lineno, trace = Checker(net, options.progress).run()
        if options.progress :
            print("%s states explored" % states)
    if options.simul :
        engine = "dot"
        for eng in gv_engines :
            if getattr(options, "gv%s" % eng) :
                engine = eng
                break
        try :
            simul = Simulator(node, net, draw(net, None, engine), options.port)
        except :
            bug()
        simul.start()
        if options.headless :
            with open(options.headless, "w") as out :
                out.write(json({"res" : "%sr" % simul.url,
                                "url" : simul.url,
                                "key" : simul.server.httpd.key,
                                "host" : "127.0.0.1",
                                "port" : simul.port}))
        else :
            webbrowser.open(simul.url)
        simul.wait()
    elif trace :
        if lineno is None :
            print("unsafe execution:")
        else :
            asserts = dict((a.lineno, a) for a in net.label("asserts"))
            print("line %s, %r failed:"
                  % (lineno, asserts[lineno].st.source()))
        for trans, mode in trace :
            print("  %s %s" % (trans, mode))
    return net

if __name__ == "__main__" :
    main()



"""This module features a parser for CTL* formula. Function `build`
parses the formula (see _concrete syntax_ below) and expands
sub-formulas; then it returns an abstract syntax tree (see _abstract
syntax_) below.

>>> from snakes.utils.ctlstar.build import *
>>> spec = '''
... atom even (x : int) :
...     return x % 2 == 0
... atom odd (x : int) :
...     return x % 2 == 1
... prop evenplace (p : place) :
...     exists tok in p (even(x=tok))
... prop oddplace (p : place) :
...     exists tok in p (odd(x=tok))
... A (G (evenplace(p=@'some_place') => F oddplace(p=@'another_place')))
... '''
>>> tree = build(spec)
>>> print ast.dump(tree)
CtlUnary(op=All(), child=CtlUnary(op=Globally(), child=CtlBinary(op=Imply(), ...

In the example above, `atom` allows to define parameterised atomic
propositions and `prop` allows to define sub-formulas. In the returned
syntax tree, all these objects are inlined and fully instantiated.

## Concrete syntax ##
"""

# apidoc include "../../lang/ctlstar/ctlstar.pgen" first=6 last=35 lang="text"

"""
## Abstract syntax ##
"""

# apidoc include "../../lang/ctlstar/ctlstar.asdl" first=3 last=59 lang="text"
pass



from snakes.utils.simul import BaseSimulator, BaseHTTPSimulator
import snakes.utils.abcd.html as html

class ABCDSimulator (BaseSimulator) :
    def __init__ (self, node, net, gv) :
        BaseSimulator.__init__(self, net)
        a2html = html.ABCD2HTML(node)
        n2html = html.Net2HTML(net, gv, a2html)
        self.info = {"filename" :  node.st.filename,
                     "abcd" : a2html.html(),
                     "tree" : n2html.html(),
                     "net" : n2html.svg()}
        self.tree = {}
        for node in net.node() :
            nid = gv.nodemap[node.name]
            if nid in n2html.n2t :
                self.tree[node.name] = "#" + n2html.n2t[nid]
        self.places = [place.name for place in net.place()
                       if place.name in self.tree]
        self.abcd = {}
        self.transid = []
        for trans in net.transition() :
            nid = gv.nodemap[trans.name]
            self.transid.append(self.tree[trans.name])
            if nid in n2html.n2a :
                self.abcd[trans.name] = ", ".join("#" + i for i in
                                                  n2html.n2a[nid])
    def init (self, state=-1) :
        res = BaseSimulator.init(self, state)
        res.update(self.info)
        res["help"] = self.init_help()
        return res
    def init_help (self) :
        help = BaseSimulator.init_help(self)
        help.update({
            "#model .abcd" : {
                "title" : "Source",
                "content" : "ABCD source code"
            },
            "#model .tree" : {
                "title" : "State",
                "content" : "hierarchy of ABCD objects",
            },
            "#model .petrinet" : {
                "title" : "Net",
                "content" : "Petri nets semantics"
            }})
        return help
    def getstate (self, state) :
        marking = self.states[state]
        modes = dict((t, []) for t in self.transid)
        for i, (trans, mode) in enumerate(marking.modes) :
            modes[self.tree[trans.name]].append({"state" : state,
                                                 "mode" : i,
                                                 "html" : str(mode)})
        return {"id" : state,
                "states" :
                [{"do" : "dropclass",
                  "select" : "#model .active",
                  "class" : "active"}]
                + [{"do" : "addclass",
                    "select" : self.abcd[trans.name],
                    "class" : "active"} for trans, mode in marking.modes]
                + [{"do" : "addclass",
                    "select" : self.tree[trans.name],
                    "class" : "active"} for trans, mode in marking.modes]
                + [{"do" : "settext",
                    "select" : "%s .content" % self.tree[place],
                    "text" : "{}"} for place in self.places
                   if place not in marking]
                + [{"do" : "settext",
                    "select" : "%s .content" % self.tree[place],
                    "text" : str(marking[place])} for place in marking
                   if place in self.tree],
                "modes" : [{"select" : "%s + .modes" % trans,
                            "items" : items}
                           for trans, items in modes.items()],
                }

class Simulator (BaseHTTPSimulator) :
    def __init__ (self, node, net, gv, port) :
        simul = ABCDSimulator(node, net, gv)
        BaseHTTPSimulator.__init__(self, net, simulator=simul, port=port)
    def init_model (self) :
        return self.res["model.html"] % self.simul.info
    def init_ui (self) :
        return BaseHTTPSimulator.init_ui(self)[:-1] + [{
            "label" : "Show net",
            "id" : "ui-shownet",
            "href" : "#",
            "script" : "dialog($('#model .petrinet').html())"
            }]



"""
>>> testparser(Translator)
"""

import doctest, traceback, sys, re, operator, inspect
from snakes.lang.pgen import ParseError
from snakes.lang.python.pgen import parser
import snakes.lang.python.asdl as ast
from snakes import *

_symbols = parser.tokenizer.tok_name.copy()
# next statement overrides 'NT_OFFSET' entry with 'single_input'
# (this is desired)
_symbols.update(parser.symbolMap)

class ParseTree (list) :
    _symbols = _symbols
    def __init__ (self, st) :
        self.symbol = self._symbols[st[0][0]]
        self.kind = st[0][0]
        self.srow = st[0][1].srow
        self.erow = st[0][1].erow
        self.scol = st[0][1].scol
        self.ecol = st[0][1].ecol
        self.text = st[0][1]
        self.filename = self.text.filename
        list.__init__(self, (self.__class__(child) for child in st[1]))
    def __repr__ (self) :
        return repr(self.text)
    def involve (self, rule) :
        if self.kind == rule :
            return True
        for child in self :
            if child.involve(rule) :
                return True
        return False
    def source (self) :
        lines = self.text.lexer.lines[self.srow-1:self.erow]
        if self.srow == self.erow :
            lines[0] = lines[0][self.scol:self.ecol]
        else :
            lines[0] = lines[0][self.scol:]
            lines[-1] = lines[-1][:self.ecol]
        return str("\n".join(l.rstrip("\n") for l in lines))

class Translator (object) :
    ParseTree = ParseTree
    parser = parser
    ST = ast
    def __init__ (self, st) :
        for value, name in self.parser.tokenizer.tok_name.items() :
            setattr(self, name, value)
        def isdict (obj) : return isinstance(obj, dict)
        for name, d in inspect.getmembers(self.__class__, isdict) :
            d = d.copy()
            for key, val in d.items() :
                try :
                    d[key] = getattr(self.ST, val.__name__)
                except AttributeError :
                    pass
            setattr(self, name, d)
        self.ast = self.do(self.ParseTree(st))
    def do (self, st, ctx=None) :
        if ctx is None :
            ctx = self.ST.Load
        name = st.symbol
        meth = getattr(self, "do_" + name)
        tree = meth(st, ctx)
        try :
            tree.st = st
        except AttributeError :
            pass
        return tree
    # unary operations
    _unary = {"not" : ast.Not,
              "+" : ast.UAdd,
              "-" : ast.USub,
              "~" : ast.Invert}
    def _do_unary (self, st, ctx) :
        """unary: not_test | factor
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.UnaryOp(lineno=st.srow, col_offset=st.scol,
                                   op=self._unary[st[0].text](lineno=st[0].srow,
                                                              col_offset=st[0].scol),
                                   operand=self.do(st[1], ctx))
    # binary operations
    _binary = {"+" : ast.Add,
               "-" : ast.Sub,
               "*" : ast.Mult,
               "/" : ast.Div,
               "%" : ast.Mod,
               "&" : ast.BitAnd,
               "|" : ast.BitOr,
               "^" : ast.BitXor,
               "<<" : ast.LShift,
               ">>" : ast.RShift,
               "**" : ast.Pow,
               "//" : ast.FloorDiv}
    def _do_binary (self, st, ctx) :
        """binary: expr | xor_expr | and_expr | shift_expr | arith_expr | term
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            values = [self.do(child, ctx) for child in st[::2]]
            ops = [(self._binary[child.text], child) for child in st[1::2]]
            while len(values) > 1 :
                left = values.pop(0)
                right = values.pop(0)
                operator, node = ops.pop(0)
                values.insert(0, self.ST.BinOp(lineno=st.srow,
                                               col_offset=st.scol,
                                               left=left,
                                               op=operator(lineno=node.srow,
                                                           col_offset=node.scol),
                                               right=right))
            return values[0]
    # boolean operation
    _boolean = {"and" : ast.And,
                "or" : ast.Or}
    def _do_boolean (self, st, ctx) :
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.BoolOp(lineno=st.srow, col_offset=st.scol,
                                  op=self._boolean[st[1].text](lineno=st[1].srow,
                                                               col_offset=st[1].scol),
                                  values=[self.do(child, ctx)
                                          for child in st[::2]])
    # start of rule handlers
    def do_file_input (self, st, ctx) :
        """file_input: (NEWLINE | stmt)* ENDMARKER
        -> ast.Module

        <<< pass
        'Module(body=[Pass()])'
        """
        body = reduce(operator.add,
                      (self.do(child, ctx)  for child in st
                       if child.kind not in (self.NEWLINE, self.ENDMARKER)),
                      [])
        return self.ST.Module(lineno=st.srow,
                              col_offset=st.scol,
                              body=body)
    def do_decorator (self, st, ctx) :
        """decorator: '@' dotted_name [ '(' [arglist] ')' ] NEWLINE
        -> ast.AST

        <<< @foo
        ... def f() : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Name(id='foo', ctx=Load())], returns=None)])"
        <<< @foo.bar
        ... def f() : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Attribute(value=Name(id='foo', ctx=Load()), attr='bar', ctx=Load())], returns=None)])"
        <<< @foo()
        ... def f() : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Call(func=Name(id='foo', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None)], returns=None)])"
        <<< @foo(x, y)
        ... def f() : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Call(func=Name(id='foo', ctx=Load()), args=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())], keywords=[], starargs=None, kwargs=None)], returns=None)])"
        <<< @foo.bar(x, y)
        ... def f() : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Call(func=Attribute(value=Name(id='foo', ctx=Load()), attr='bar', ctx=Load()), args=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())], keywords=[], starargs=None, kwargs=None)], returns=None)])"
        """
        names = self.do(st[1], ctx).split(".")
        obj = self.ST.Name(lineno=st[1].srow, col_offset=st[1].scol,
                           id=names.pop(0), ctx=ctx())
        while names :
            obj = self.ST.Attribute(lineno=st[1].srow,
                                    col_offset=st[1].scol,
                                    value=obj, attr=names.pop(0), ctx=ctx())
        if len(st) == 3 :
            return obj
        elif len(st) == 5 :
            return self.ST.Call(lineno=st[1].srow, col_offset=st[1].scol,
                                func=obj, args=[], keywords=[],
                                starargs=None, kwargs=None)
        else :
            args, keywords, starargs, kwargs = self.do(st[3], ctx)
            return self.ST.Call(lineno=st[1].srow, col_offset=st[1].scol,
                                func=obj, args=args, keywords=keywords,
                                starargs=starargs, kwargs=kwargs)
    def do_decorators (self, st, ctx) :
        """decorators: decorator+
        -> ast.AST+

        <<< @foo
        ... @bar
        ... @spam.egg()
        ... def f () :
        ...     pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Name(id='foo', ctx=Load()), Name(id='bar', ctx=Load()), Call(func=Attribute(value=Name(id='spam', ctx=Load()), attr='egg', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None)], returns=None)])"
        """
        return [self.do(child, ctx) for child in st]
    def do_decorated (self, st, ctx) :
        """decorated: decorators (classdef | funcdef)
        -> ast.AST

        <<< @foo
        ... def f () :
        ...     pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Name(id='foo', ctx=Load())], returns=None)])"
        <<< @foo
        ... class c () :
        ...     pass
        "Module(body=[ClassDef(name='c', bases=[], keywords=[], starargs=None, kwargs=None, body=[Pass()], decorator_list=[Name(id='foo', ctx=Load())])])"
        """
        child = self.do(st[1], ctx)
        child.decorator_list.extend(self.do(st[0], ctx))
        return child
    def do_funcdef (self, st, ctx) :
        """funcdef: 'def' NAME parameters ['->' test] ':' suite
        -> ast.FunctionDef

        <<< def f(x, y) : x+y
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Expr(value=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())))], decorator_list=[], returns=None)])"
        <<< def f(x, y) -> int : x+y
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Expr(value=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())))], decorator_list=[], returns=Name(id='int', ctx=Load()))])"
        """
        if len(st) == 5 :
            return self.ST.FunctionDef(lineno=st.srow, col_offset=st.scol,
                                       name=st[1].text,
                                       args=self.do(st[2], ctx),
                                       returns=None,
                                       body=self.do(st[-1], ctx),
                                       decorator_list=[])
        else :
            return self.ST.FunctionDef(lineno=st.srow, col_offset=st.scol,
                                       name=st[1].text,
                                       args=self.do(st[2], ctx),
                                       returns=self.do(st[5], ctx),
                                       body=self.do(st[-1], ctx),
                                       decorator_list=[])
    def do_parameters (self, st, ctx) :
        """parameters: '(' [typedargslist] ')'
        -> ast.arguments

        <<< def f () : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x, y) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        """
        if len(st) == 2 :
            return self.ST.arguments(lineno=st.srow, col_offset=st.scol,
                                     args=[], vararg=None,
                                     varargannotation=None,
                                     kwonlyargs=[], kwarg=None,
                                     kwargannotation=None,
                                     defaults=[], kw_defaults=[])
        else :
            return self.do(st[1], ctx)
    def do_typedargslist (self, st, ctx) :
        """typedargslist: ((tfpdef ['=' test] ',')*
                ('*' [tfpdef] (',' tfpdef ['=' test])* [',' '**' tfpdef]
                 | '**' tfpdef)
                | tfpdef ['=' test] (',' tfpdef ['=' test])* [','])
        -> ast.arguments

        <<< def f(x) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x, *y, z) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None)], vararg='y', varargannotation=None, kwonlyargs=[arg(arg='z', annotation=None)], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[None]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(*, x=1) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[arg(arg='x', annotation=None)], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[Num(n=1)]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x, y, *z, a=1, b=2, **d) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg='z', varargannotation=None, kwonlyargs=[arg(arg='a', annotation=None), arg(arg='b', annotation=None)], kwarg='d', kwargannotation=None, defaults=[], kw_defaults=[Num(n=1), Num(n=2)]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x:int) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=Name(id='int', ctx=Load()))], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x:int, *y:float, z:bool) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=Name(id='int', ctx=Load()))], vararg='y', varargannotation=Name(id='float', ctx=Load()), kwonlyargs=[arg(arg='z', annotation=Name(id='bool', ctx=Load()))], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[None]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(*, x:int=1) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[arg(arg='x', annotation=Name(id='int', ctx=Load()))], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[Num(n=1)]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x:int, y, *z:float, a=1, b:bool=False, **d:object) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=Name(id='int', ctx=Load())), arg(arg='y', annotation=None)], vararg='z', varargannotation=Name(id='float', ctx=Load()), kwonlyargs=[arg(arg='a', annotation=None), arg(arg='b', annotation=Name(id='bool', ctx=Load()))], kwarg='d', kwargannotation=Name(id='object', ctx=Load()), defaults=[], kw_defaults=[Num(n=1), Name(id='False', ctx=Load())]), body=[Pass()], decorator_list=[], returns=None)])"
        """
        args = []
        vararg = None
        varargannotation = None
        star = False
        kwonlyargs = []
        kwarg = None
        kwargannotation = None
        defaults = []
        kw_defaults = []
        nodes = list(st)
        while nodes :
            first = nodes.pop(0)
            if first.text == "," :
                pass
            elif first.text == "*" :
                star = True
                if nodes and nodes[0].text != "," :
                    vararg, varargannotation = self.do(nodes.pop(0), ctx)
            elif first.text == "**" :
                kwarg, kwargannotation = self.do(nodes.pop(0), ctx)
            else :
                n, a = self.do(first, ctx)
                arg = self.ST.arg(lineno=first.srow, col_offset=first.scol,
                                  arg=n, annotation=a)
                if nodes and nodes[0].text == "=" :
                    del nodes[0]
                    d = self.do(nodes.pop(0), ctx)
                else :
                    d = None
                if star :
                    kwonlyargs.append(arg)
                    kw_defaults.append(d)
                else :
                    args.append(arg)
                    if d is not None :
                        defaults.append(d)
        return self.ST.arguments(lineno=st.srow, col_offset=st.scol,
                                 args=args,
                                 vararg=vararg,
                                 varargannotation=varargannotation,
                                 kwonlyargs=kwonlyargs,
                                 kwarg=kwarg,
                                 kwargannotation=kwargannotation,
                                 defaults=defaults,
                                 kw_defaults=kw_defaults)
    def do_tfpdef (self, st, ctx) :
        """tfpdef: NAME [':' test]
        -> str, ast.AST?

        <<< def f(x:int) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=Name(id='int', ctx=Load()))], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x:int, y:float) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=Name(id='int', ctx=Load())), arg(arg='y', annotation=Name(id='float', ctx=Load()))], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< def f(x, y:int) : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=Name(id='int', ctx=Load()))], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        """
        if len(st) == 1 :
            return st[0].text, None
        else :
            return st[0].text, self.do(st[2], ctx)
    def do_varargslist (self, st, ctx) :
        """varargslist: ((vfpdef ['=' test] ',')*
              ('*' [vfpdef] (',' vfpdef ['=' test])*  [',' '**' vfpdef]
               | '**' vfpdef)
              | vfpdef ['=' test] (',' vfpdef ['=' test])* [','])
        -> ast.arguments

        <<< lambda x, y : x+y
        "Module(body=[Expr(value=Lambda(args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load()))))])"
        <<< lambda x, y, z=1 : x+y+z
        "Module(body=[Expr(value=Lambda(args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None), arg(arg='z', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[Num(n=1)], kw_defaults=[]), body=BinOp(left=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())), op=Add(), right=Name(id='z', ctx=Load()))))])"
        """
        tree = self.do_typedargslist(st, ctx)
        tree.st = st
        return tree
    def do_vfpdef (self, st, ctx) :
        """vfpdef: NAME
        -> str, None

        Return value (str, None) is choosen for compatibility with
        do_tfpdef, so that do_typedargslist can be used for
        do_varargslist.

        <<< lambda x, y : x+y
        "Module(body=[Expr(value=Lambda(args=arguments(args=[arg(arg='x', annotation=None), arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load()))))])"
        """
        return st[0].text, None
    def do_stmt (self, st, ctx) :
        """stmt: simple_stmt | compound_stmt
        -> ast.AST+

        <<< pass
        'Module(body=[Pass()])'
        <<< pass; pass
        'Module(body=[Pass(), Pass()])'
        <<< with x : y
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Expr(value=Name(id='y', ctx=Load()))])])"
        """
        child = self.do(st[0], ctx)
        if isinstance(child, self.ST.AST) :
            return [child]
        else :
            return child
    def do_simple_stmt (self, st, ctx) :
        """simple_stmt: small_stmt (';' small_stmt)* [';'] NEWLINE
        -> ast.AST+

        <<< del x; pass; import spam as egg
        "Module(body=[Delete(targets=[Name(id='x', ctx=Del())]), Pass(), Import(names=[alias(name='spam', asname='egg')])])"
        """
        return [self.do(child, ctx) for child in st[::2]
                if child.kind != self.NEWLINE]
    def do_small_stmt (self, st, ctx) :
        """small_stmt: (expr_stmt | del_stmt | pass_stmt | flow_stmt |
                     import_stmt | global_stmt | nonlocal_stmt | assert_stmt)
        -> ast.AST

        <<< x
        "Module(body=[Expr(value=Name(id='x', ctx=Load()))])"
        <<< del x
        "Module(body=[Delete(targets=[Name(id='x', ctx=Del())])])"
        <<< pass
        'Module(body=[Pass()])'
        <<< import egg as spam
        "Module(body=[Import(names=[alias(name='egg', asname='spam')])])"
        <<< global x
        "Module(body=[Global(names=['x'])])"
        <<< nonlocal x
        "Module(body=[Nonlocal(names=['x'])])"
        <<< assert x
        "Module(body=[Assert(test=Name(id='x', ctx=Load()), msg=None)])"
        """
        return self.do(st[0], ctx)
    def do_expr_stmt (self, st, ctx) :
        """expr_stmt: testlist (augassign (yield_expr|testlist) |
                                ('=' (yield_expr|testlist))*)
        -> ast.Expr

        <<< x = y = 1
        "Module(body=[Assign(targets=[Name(id='x', ctx=Store()), Name(id='y', ctx=Store())], value=Num(n=1))])"
        <<< x, y = z = 1, 2
        "Module(body=[Assign(targets=[Tuple(elts=[Name(id='x', ctx=Store()), Name(id='y', ctx=Store())], ctx=Store()), Name(id='z', ctx=Store())], value=Tuple(elts=[Num(n=1), Num(n=2)], ctx=Load()))])"
        <<< x += 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Add(), value=Num(n=1))])"
        <<< x += yield 5
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Add(), value=Yield(value=Num(n=5)))])"
        <<< x = y = yield 1, 2
        "Module(body=[Assign(targets=[Name(id='x', ctx=Store()), Name(id='y', ctx=Store())], value=Yield(value=Tuple(elts=[Num(n=1), Num(n=2)], ctx=Load())))])"
        """
        if len(st) == 1 :
            return self.ST.Expr(lineno=st.srow, col_offset=st.scol,
                                value=self.do(st[0], ctx))
        elif st[1].symbol == "augassign" :
            target = self.do(st[0], self.ST.Store)
            if isinstance(target, self.ST.Tuple) :
                raise ParseError(st[0].text, reason="illegal expression for"
                                 " augmented assignment")
            return self.ST.AugAssign(lineno=st.srow, col_offset=st.scol,
                                     target=target,
                                     op=self.do(st[1], ctx),
                                     value=self.do(st[2], ctx))
        else :
            return self.ST.Assign(lineno=st.srow, col_offset=st.scol,
                                  targets=[self.do(child, ast.Store)
                                           for child in st[:-1:2]],
                                  value=self.do(st[-1], ctx))
    def do_augassign (self, st, ctx) :
        """augassign: ('+=' | '-=' | '*=' | '/=' | '%=' | '&=' | '|=' | '^=' |
            '<<=' | '>>=' | '**=' | '//=')
        -> ast.AST

        <<< x += 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Add(), value=Num(n=1))])"
        <<< x -= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Sub(), value=Num(n=1))])"
        <<< x *= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Mult(), value=Num(n=1))])"
        <<< x /= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Div(), value=Num(n=1))])"
        <<< x %= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Mod(), value=Num(n=1))])"
        <<< x &= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=BitAnd(), value=Num(n=1))])"
        <<< x |= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=BitOr(), value=Num(n=1))])"
        <<< x ^= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=BitXor(), value=Num(n=1))])"
        <<< x <<= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=LShift(), value=Num(n=1))])"
        <<< x >>= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=RShift(), value=Num(n=1))])"
        <<< x **= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=Pow(), value=Num(n=1))])"
        <<< x //= 1
        "Module(body=[AugAssign(target=Name(id='x', ctx=Store()), op=FloorDiv(), value=Num(n=1))])"
        """
        return self._binary[st[0].text[:-1]](lineno=st.srow, col_offset=st.scol)
    def do_del_stmt (self, st, ctx) :
        """del_stmt: 'del' exprlist
        -> ast.Delete

        <<< del x
        "Module(body=[Delete(targets=[Name(id='x', ctx=Del())])])"
        <<< del x, y
        "Module(body=[Delete(targets=[Name(id='x', ctx=Del()), Name(id='y', ctx=Del())])])"
        """
        targets = self.do(st[1], ctx=self.ST.Del)
        if isinstance(targets, self.ST.Tuple) :
            targets = targets.elts
        else :
            targets = [targets]
        return self.ST.Delete(lineno=st.srow, col_offset=st.scol,
                              targets=targets)
    def do_pass_stmt (self, st, ctx) :
        """pass_stmt: 'pass'
        -> ast.Pass

        <<< pass
        'Module(body=[Pass()])'
        """
        return self.ST.Pass(lineno=st.srow, col_offset=st.scol)
    def do_flow_stmt (self, st, ctx) :
        """flow_stmt: break_stmt | continue_stmt | return_stmt | raise_stmt
                 | yield_stmt
        -> ast.AST

        <<< break
        'Module(body=[Break()])'
        <<< continue
        'Module(body=[Continue()])'
        <<< return
        'Module(body=[Return(value=None)])'
        <<< raise
        'Module(body=[Raise(exc=None, cause=None)])'
        <<< yield
        'Module(body=[Expr(value=Yield(value=None))])'
        """
        return self.do(st[0], ctx)
    def do_break_stmt (self, st, ctx) :
        """break_stmt: 'break'
        -> ast.Break()

        <<< break
        'Module(body=[Break()])'
        """
        return self.ST.Break(lineno=st.srow, col_offset=st.scol)
    def do_continue_stmt (self, st, ctx) :
        """continue_stmt: 'continue'
        -> ast.Continue

        <<< continue
        'Module(body=[Continue()])'
        """
        return self.ST.Continue(lineno=st.srow, col_offset=st.scol)
    def do_return_stmt (self, st, ctx) :
        """return_stmt: 'return' [testlist]
        -> ast.Return

        <<< return
        'Module(body=[Return(value=None)])'
        <<< return 1
        'Module(body=[Return(value=Num(n=1))])'
        <<< return 1, 2
        'Module(body=[Return(value=Tuple(elts=[Num(n=1), Num(n=2)], ctx=Load()))])'
        """
        if len(st) == 1 :
            return self.ST.Return(lineno=st.srow, col_offset=st.scol,
                                  value=None)

        else :
            return self.ST.Return(lineno=st.srow, col_offset=st.scol,
                                  value=self.do(st[1], ctx))
    def do_yield_stmt (self, st, ctx) :
        """yield_stmt: yield_expr
        -> ast.Expr

        <<< yield
        'Module(body=[Expr(value=Yield(value=None))])'
        <<< yield 42
        'Module(body=[Expr(value=Yield(value=Num(n=42)))])'
        """
        return self.ST.Expr(lineno=st.srow, col_offset=st.scol,
                            value=self.do(st[0], ctx))
    def do_raise_stmt (self, st, ctx) :
        """raise_stmt: 'raise' [test ['from' test]]
        -> ast.Raise

        <<< raise
        'Module(body=[Raise(exc=None, cause=None)])'
        <<< raise Exception
        "Module(body=[Raise(exc=Name(id='Exception', ctx=Load()), cause=None)])"
        <<< raise Exception from Exception
        "Module(body=[Raise(exc=Name(id='Exception', ctx=Load()), cause=Name(id='Exception', ctx=Load()))])"
        """
        count = len(st)
        if count == 1 :
            return self.ST.Raise(lineno=st.srow, col_offset=st.scol,
                                 exc=None, cause=None)
        elif count == 2 :
            return self.ST.Raise(lineno=st.srow, col_offset=st.scol,
                                 exc=self.do(st[1], ctx),
                                 cause=None)
        else :
            return self.ST.Raise(lineno=st.srow, col_offset=st.scol,
                                 exc=self.do(st[1], ctx),
                                 cause=self.do(st[3], ctx))
    def do_import_stmt (self, st, ctx) :
        """import_stmt: import_name | import_from
        -> ast.AST

        <<< import foo
        "Module(body=[Import(names=[alias(name='foo', asname=None)])])"
        <<< import foo.bar
        "Module(body=[Import(names=[alias(name='foo.bar', asname=None)])])"
        <<< import foo as bar
        "Module(body=[Import(names=[alias(name='foo', asname='bar')])])"
        <<< import foo.bar as egg
        "Module(body=[Import(names=[alias(name='foo.bar', asname='egg')])])"
        <<< import foo as bar, egg as spam
        "Module(body=[Import(names=[alias(name='foo', asname='bar'), alias(name='egg', asname='spam')])])"
        <<< import foo, bar, egg as spam
        "Module(body=[Import(names=[alias(name='foo', asname=None), alias(name='bar', asname=None), alias(name='egg', asname='spam')])])"
        """
        return self.do(st[0], ctx)
    def do_import_name (self, st, ctx) :
        """import_name: 'import' dotted_as_names
        -> ast.Import

        <<< import foo
        "Module(body=[Import(names=[alias(name='foo', asname=None)])])"
        <<< import foo.bar
        "Module(body=[Import(names=[alias(name='foo.bar', asname=None)])])"
        <<< import foo as bar
        "Module(body=[Import(names=[alias(name='foo', asname='bar')])])"
        <<< import foo.bar as egg
        "Module(body=[Import(names=[alias(name='foo.bar', asname='egg')])])"
        <<< import foo as bar, egg as spam
        "Module(body=[Import(names=[alias(name='foo', asname='bar'), alias(name='egg', asname='spam')])])"
        <<< import foo, bar, egg as spam
        "Module(body=[Import(names=[alias(name='foo', asname=None), alias(name='bar', asname=None), alias(name='egg', asname='spam')])])"
        """
        return self.ST.Import(lineno=st.srow, col_offset=st.scol,
                              names=self.do(st[1], ctx))
    def do_import_from (self, st, ctx) :
        """import_from: ('from' (('.' | '...')* dotted_name | ('.' | '...')+)
              'import' ('*' | '(' import_as_names ')' | import_as_names))
        -> ast.ImportFrom

        <<< from foo import egg as spam
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam')], level=0)])"
        <<< from .foo import egg as spam
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam')], level=1)])"
        <<< from ...foo import egg as spam
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam')], level=3)])"
        <<< from ...foo.bar import egg as spam
        "Module(body=[ImportFrom(module='foo.bar', names=[alias(name='egg', asname='spam')], level=3)])"
        <<< from foo import egg as spam, bar as baz
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam'), alias(name='bar', asname='baz')], level=0)])"
        <<< from .foo import egg as spam, bar as baz
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam'), alias(name='bar', asname='baz')], level=1)])"
        <<< from ...foo import egg as spam, bar as baz
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam'), alias(name='bar', asname='baz')], level=3)])"
        <<< from ...foo.bar import egg as spam, baz
        "Module(body=[ImportFrom(module='foo.bar', names=[alias(name='egg', asname='spam'), alias(name='baz', asname=None)], level=3)])"
        <<< from foo import *
        "Module(body=[ImportFrom(module='foo', names=[alias(name='*', asname=None)], level=0)])"
        <<< from .foo import *
        "Module(body=[ImportFrom(module='foo', names=[alias(name='*', asname=None)], level=1)])"
        <<< from ...foo import *
        "Module(body=[ImportFrom(module='foo', names=[alias(name='*', asname=None)], level=3)])"
        <<< from ...foo.bar import *
        "Module(body=[ImportFrom(module='foo.bar', names=[alias(name='*', asname=None)], level=3)])"
        """
        level = 0
        next = 1
        for i, child in enumerate(st[1:]) :
            text = child.text
            if text not in (".", "...") :
                next = i+1
                break
            level += len(text)
        if text == "import" :
            module = ""
            next += 1
        else :
            module = self.do(st[next], ctx)
            next += 2
        text = st[next].text
        if text == "*" :
            names = [self.ST.alias(lineno=st[next].srow,
                                   col_offset=st[next].scol,
                                   name="*", asname=None)]
        elif text == "(" :
            names = self.do(st[next+1], ctx)
        else :
            names = self.do(st[next], ctx)
        return self.ST.ImportFrom(lineno=st.srow, col_offset=st.scol,
                                  module=module, names=names, level=level)
    def do_import_as_name (self, st, ctx) :
        """import_as_name: NAME ['as' NAME]
        -> ast.alias

        <<< from foo import egg as spam
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam')], level=0)])"
        """
        if len(st) == 1 :
            return self.ST.alias(lineno=st.srow, col_offset=st.scol,
                                 name=st[0].text,
                                 asname=None)
        else :
            return self.ST.alias(lineno=st.srow, col_offset=st.scol,
                                 name=st[0].text,
                                 asname=st[2].text)
    def do_dotted_as_name (self, st, ctx) :
        """dotted_as_name: dotted_name ['as' NAME]
        -> ast.alias

        <<< import foo.bar as egg
        "Module(body=[Import(names=[alias(name='foo.bar', asname='egg')])])"
        """
        if len(st) == 1 :
            return self.ST.alias(lineno=st.srow, col_offset=st.scol,
                                 name=self.do(st[0], ctx),
                                 asname=None)
        else :
            return self.ST.alias(lineno=st.srow, col_offset=st.scol,
                                 name=self.do(st[0], ctx),
                                 asname=st[2].text)
    def do_import_as_names (self, st, ctx) :
        """import_as_names: import_as_name (',' import_as_name)* [',']
        -> ast.alias+

        <<< from foo import egg as spam
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam')], level=0)])"
        <<< from foo import egg as spam, bar as baz
        "Module(body=[ImportFrom(module='foo', names=[alias(name='egg', asname='spam'), alias(name='bar', asname='baz')], level=0)])"
        """
        return [self.do(child, ctx) for child in st[::2]]
    def do_dotted_as_names (self, st, ctx) :
        """dotted_as_names: dotted_as_name (',' dotted_as_name)*
        -> ast.alias+

        <<< import foo.bar, egg.spam as baz
        "Module(body=[Import(names=[alias(name='foo.bar', asname=None), alias(name='egg.spam', asname='baz')])])"
        """
        return [self.do(child, ctx) for child in st[::2]]
    def do_dotted_name (self, st, ctx) :
        """dotted_name: NAME ('.' NAME)*
        -> str

        <<< import foo.bar
        "Module(body=[Import(names=[alias(name='foo.bar', asname=None)])])"
        """
        return ".".join(child.text for child in st[::2])
    def do_global_stmt (self, st, ctx) :
        """global_stmt: 'global' NAME (',' NAME)*
        -> ast.Global

        <<< global x
        "Module(body=[Global(names=['x'])])"
        <<< global x, y
        "Module(body=[Global(names=['x', 'y'])])"
        """
        return self.ST.Global(lineno=st.srow, col_offset=st.scol,
                              names=[child.text for child in st[1::2]])
    def do_nonlocal_stmt (self, st, ctx) :
        """nonlocal_stmt: 'nonlocal' NAME (',' NAME)*
        -> ast.Nonlocal

        <<< nonlocal x
        "Module(body=[Nonlocal(names=['x'])])"
        <<< nonlocal x, y
        "Module(body=[Nonlocal(names=['x', 'y'])])"
        """
        return self.ST.Nonlocal(lineno=st.srow, col_offset=st.scol,
                                names=[child.text for child in st[1::2]])
    def do_assert_stmt (self, st, ctx) :
        """assert_stmt: 'assert' test [',' test]
        -> ast.Assert

        <<< assert x
        "Module(body=[Assert(test=Name(id='x', ctx=Load()), msg=None)])"
        <<< assert x, y
        "Module(body=[Assert(test=Name(id='x', ctx=Load()), msg=Name(id='y', ctx=Load()))])"
        """
        if len(st) == 2 :
            return self.ST.Assert(lineno=st.srow, col_offset=st.scol,
                                  test=self.do(st[1], ctx),
                                  msg=None)
        else :
            return self.ST.Assert(lineno=st.srow, col_offset=st.scol,
                                  test=self.do(st[1], ctx),
                                  msg=self.do(st[3], ctx))
    def do_compound_stmt (self, st, ctx) :
        """compound_stmt: (if_stmt | while_stmt | for_stmt | try_stmt
                | with_stmt | funcdef | classdef | decorated)
        -> ast.AST

        <<< with x : pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Pass()])])"
        <<< if x : pass
        "Module(body=[If(test=Name(id='x', ctx=Load()), body=[Pass()], orelse=[])])"
        <<< while x : pass
        "Module(body=[While(test=Name(id='x', ctx=Load()), body=[Pass()], orelse=[])])"
        <<< for x in l : pass
        "Module(body=[For(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), body=[Pass()], orelse=[])])"
        <<< try : pass
        ... except : pass
        'Module(body=[TryExcept(body=[Pass()], handlers=[ExceptHandler(type=None, name=None, body=[Pass()])], orelse=[])])'
        <<< def f () : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[], returns=None)])"
        <<< class c () : pass
        "Module(body=[ClassDef(name='c', bases=[], keywords=[], starargs=None, kwargs=None, body=[Pass()], decorator_list=[])])"
        <<< @foo
        ... def f () : pass
        "Module(body=[FunctionDef(name='f', args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=[Pass()], decorator_list=[Name(id='foo', ctx=Load())], returns=None)])"
        """
        return self.do(st[0], ctx)
    def do_if_stmt (self, st, ctx) :
        """if_stmt: 'if' test ':' suite ('elif' test ':' suite)*
               ['else' ':' suite]
        -> ast.If

        <<< if x == 1 : pass
        "Module(body=[If(test=Compare(left=Name(id='x', ctx=Load()), ops=[Eq()], comparators=[Num(n=1)]), body=[Pass()], orelse=[])])"
        <<< if x == 1 : pass
        ... else : pass
        "Module(body=[If(test=Compare(left=Name(id='x', ctx=Load()), ops=[Eq()], comparators=[Num(n=1)]), body=[Pass()], orelse=[Pass()])])"
        <<< if x == 1 : pass
        ... elif y == 2 : pass
        ... elif z == 3 : pass
        ... else : pass
        "Module(body=[If(test=Compare(left=Name(id='x', ctx=Load()), ops=[Eq()], comparators=[Num(n=1)]), body=[Pass()], orelse=[If(test=Compare(left=Name(id='y', ctx=Load()), ops=[Eq()], comparators=[Num(n=2)]), body=[Pass()], orelse=[If(test=Compare(left=Name(id='z', ctx=Load()), ops=[Eq()], comparators=[Num(n=3)]), body=[Pass()], orelse=[Pass()])])])])"

        """
        nodes = list(st)
        first = None
        last = None
        while nodes :
            if len(nodes) == 3 :
                last.orelse.extend(self.do(nodes[2], ctx))
                del nodes[:3]
            else :
                next = self.ST.If(lineno=nodes[0].srow,
                                  col_offset=nodes[0].scol,
                                  test=self.do(nodes[1], ctx),
                                  body=self.do(nodes[3], ctx),
                                  orelse=[])
                if first is None :
                    first = next
                if last is not None :
                    last.orelse.append(next)
                last = next
                del nodes[:4]
        return first
    def do_while_stmt (self, st, ctx) :
        """while_stmt: 'while' test ':' suite ['else' ':' suite]
        -> ast.While

        <<< while x : pass
        "Module(body=[While(test=Name(id='x', ctx=Load()), body=[Pass()], orelse=[])])"
        <<< while x :
        ...     pass
        ...     pass
        "Module(body=[While(test=Name(id='x', ctx=Load()), body=[Pass(), Pass()], orelse=[])])"
        <<< while x :
        ...     pass
        ... else :
        ...     pass
        "Module(body=[While(test=Name(id='x', ctx=Load()), body=[Pass()], orelse=[Pass()])])"
        """
        if len(st) == 4 :
            return self.ST.While(lineno=st.srow, col_offset=st.scol,
                                 test=self.do(st[1], ctx),
                                 body=self.do(st[3], ctx),
                                 orelse=[])
        else :
            return self.ST.While(lineno=st.srow, col_offset=st.scol,
                                 test=self.do(st[1], ctx),
                                 body=self.do(st[3], ctx),
                                 orelse=self.do(st[6], ctx))
    def do_for_stmt (self, st, ctx) :
        """for_stmt: 'for' exprlist 'in' testlist ':' suite ['else' ':' suite]
        -> ast.For

        <<< for x in l : pass
        "Module(body=[For(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), body=[Pass()], orelse=[])])"
        <<< for x in l :
        ...     pass
        ...     pass
        "Module(body=[For(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), body=[Pass(), Pass()], orelse=[])])"
        <<< for x in l :
        ...     pass
        ... else :
        ...     pass
        "Module(body=[For(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), body=[Pass()], orelse=[Pass()])])"
        <<< for x, y in l : pass
        "Module(body=[For(target=Tuple(elts=[Name(id='x', ctx=Store()), Name(id='y', ctx=Store())], ctx=Store()), iter=Name(id='l', ctx=Load()), body=[Pass()], orelse=[])])"
        <<< for x in a, b : pass
        "Module(body=[For(target=Name(id='x', ctx=Store()), iter=Tuple(elts=[Name(id='a', ctx=Load()), Name(id='b', ctx=Load())], ctx=Load()), body=[Pass()], orelse=[])])"
        """
        if len(st) == 6 :
            return self.ST.For(lineno=st.srow, col_offset=st.scol,
                               target=self.do(st[1], ast.Store),
                               iter=self.do(st[3], ctx),
                               body=self.do(st[5], ctx),
                               orelse=[])
        else :
            return self.ST.For(lineno=st.srow, col_offset=st.scol,
                               target=self.do(st[1], ast.Store),
                               iter=self.do(st[3], ctx),
                               body=self.do(st[5], ctx),
                               orelse=self.do(st[8], ctx))
    def do_try_stmt (self, st, ctx) :
        """try_stmt: ('try' ':' suite
           ((except_clause ':' suite)+
            ['else' ':' suite]
            ['finally' ':' suite] |
	   'finally' ':' suite))
        -> ast.TryExcept | ast.TryFinally

        <<< try : pass
        ... except : pass
        'Module(body=[TryExcept(body=[Pass()], handlers=[ExceptHandler(type=None, name=None, body=[Pass()])], orelse=[])])'
        <<< try : pass
        ... except : pass
        ... else : pass
        ... finally : pass
        'Module(body=[TryFinally(body=[TryExcept(body=[Pass()], handlers=[ExceptHandler(type=None, name=None, body=[Pass()])], orelse=[Pass()])], finalbody=[Pass()])])'
        <<< try : pass
        ... except TypeError : pass
        ... except ValueError : pass
        ... except : pass
        "Module(body=[TryExcept(body=[Pass()], handlers=[ExceptHandler(type=Name(id='TypeError', ctx=Load()), name=None, body=[Pass()]), ExceptHandler(type=Name(id='ValueError', ctx=Load()), name=None, body=[Pass()]), ExceptHandler(type=None, name=None, body=[Pass()])], orelse=[])])"
        """
        handlers = []
        finalbody = None
        orelse = []
        nodes = st[3:]
        while nodes :
            if nodes[0].text == "else" :
                orelse.extend(self.do(nodes[2], ctx))
            elif nodes[0].text == "finally" :
                finalbody = self.do(nodes[2], ctx)
            else :
                t, n = self.do(nodes[0], ctx)
                handlers.append(self.ST.ExceptHandler(lineno=nodes[0].srow,
                                                      col_offset=nodes[0].scol,
                                                      type=t, name=n,
                                                      body=self.do(nodes[2], ctx)))
            del nodes[:3]
        stmt = self.ST.TryExcept(lineno=st.srow, col_offset=st.scol,
                                 body=self.do(st[2], ctx),
                                 handlers=handlers,
                                 orelse=orelse)
        if finalbody is None :
            return stmt
        else :
            return self.ST.TryFinally(lineno=st.srow, col_offset=st.scol,
                                      body=[stmt], finalbody=finalbody)
    def do_with_stmt (self, st, ctx) :
        """with_stmt: 'with' test [ with_var ] ':' suite
        -> ast.With

        <<< with x : pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Pass()])])"
        <<< with x as y : pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=Name(id='y', ctx=Store()), body=[Pass()])])"
        <<< with x :
        ...     pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Pass()])])"
        <<< with x as y :
        ...     pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=Name(id='y', ctx=Store()), body=[Pass()])])"
        """
        if len(st) == 5 :
            return self.ST.With(lineno=st.srow, col_offset=st.scol,
                                context_expr=self.do(st[1], ctx),
                                optional_vars=self.do(st[2], self.ST.Store),
                                body=self.do(st[4], ctx))
        else :
            return self.ST.With(lineno=st.srow, col_offset=st.scol,
                                context_expr=self.do(st[1], ctx),
                                optional_vars=None,
                                body=self.do(st[3], ctx))
    def do_with_var (self, st, ctx) :
        """with_var: 'as' expr
        -> ast.Name

        <<< with x as y : pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=Name(id='y', ctx=Store()), body=[Pass()])])"
        """
        return self.do(st[1], ctx)
    def do_except_clause (self, st, ctx) :
        """except_clause: 'except' [test ['as' NAME]]
        -> ast.AST?, ast.Name?

        <<< try : pass
        ... except NameError : pass
        ... except TypeError as err : pass
        ... except : pass
        "Module(body=[TryExcept(body=[Pass()], handlers=[ExceptHandler(type=Name(id='NameError', ctx=Load()), name=None, body=[Pass()]), ExceptHandler(type=Name(id='TypeError', ctx=Load()), name=Name(id='err', ctx=Store()), body=[Pass()]), ExceptHandler(type=None, name=None, body=[Pass()])], orelse=[])])"
        """
        if len(st) == 1 :
            return None, None
        elif len(st) == 2 :
            return self.do(st[1], ctx), None
        else :
            return self.do(st[1], ctx), self.ST.Name(lineno=st[3].srow,
                                                     col_offset=st[3].scol,
                                                     id=st[3].text,
                                                     ctx=self.ST.Store())
    def do_suite (self, st, ctx) :
        """suite: simple_stmt | NEWLINE INDENT stmt+ DEDENT
        -> ast.AST+

        <<< with x : pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Pass()])])"
        <<< with x :
        ...     pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Pass()])])"
        <<< with x :
        ...     pass
        ...     pass
        "Module(body=[With(context_expr=Name(id='x', ctx=Load()), optional_vars=None, body=[Pass(), Pass()])])"
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return reduce(operator.add,
                          (self.do(child, ctx) for child in st[2:-1]),
                          [])
    def do_test (self, st, ctx) :
        """test: or_test ['if' or_test 'else' test] | lambdef
        -> ast.AST

        <<< 3
        'Module(body=[Expr(value=Num(n=3))])'
        <<< 3 if x else 4
        "Module(body=[Expr(value=IfExp(test=Name(id='x', ctx=Load()), body=Num(n=3), orelse=Num(n=4)))])"
        <<< lambda x : x+1
        "Module(body=[Expr(value=Lambda(args=arguments(args=[arg(arg='x', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Num(n=1))))])"
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.IfExp(lineno=st.srow, col_offset=st.scol,
                                 test=self.do(st[2], ctx),
                             body=self.do(st[0], ctx),
                             orelse=self.do(st[4], ctx))
    def do_test_nocond (self, st, ctx) :
        """test_nocond: or_test | lambdef_nocond
        -> ast.AST

        <<< [x for x in (lambda: True, lambda: False) if x()]
        "Module(body=[Expr(value=ListComp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Tuple(elts=[Lambda(args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=Name(id='True', ctx=Load())), Lambda(args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=Name(id='False', ctx=Load()))], ctx=Load()), ifs=[Call(func=Name(id='x', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None)])]))])"
        """
        return self.do(st[0], ctx)
    def do_lambdef (self, st, ctx) :
        """lambdef: 'lambda' [varargslist] ':' test
        -> ast.Lambda

        <<< lambda : True
        "Module(body=[Expr(value=Lambda(args=arguments(args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=Name(id='True', ctx=Load())))])"
        <<< lambda x : x+1
        "Module(body=[Expr(value=Lambda(args=arguments(args=[arg(arg='x', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Num(n=1))))])"
        """
        if len(st) == 3 :
            return self.ST.Lambda(lineno=st.srow, col_offset=st.scol,
                                  args=self.ST.arguments(lineno=st.srow,
                                                         col_offset=st.scol,
                                                         args=[], vararg=None,
                                                         varargannotation=None,
                                                         kwonlyargs=[], kwarg=None,
                                                         kwargannotation=None,
                                                         defaults=[], kw_defaults=[]),
                                  body=self.do(st[-1], ctx))
        else :
            return self.ST.Lambda(lineno=st.srow, col_offset=st.scol,
                                  args=self.do(st[1], ctx),
                                  body=self.do(st[-1], ctx))
    def do_lambdef_nocond (self, st, ctx) :
        """lambdef_nocond: 'lambda' [varargslist] ':' test_nocond
        -> ast.Lambda

        <<< [x for x in l if lambda y : x]
        "Module(body=[Expr(value=ListComp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[Lambda(args=arguments(args=[arg(arg='y', annotation=None)], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[]), body=Name(id='x', ctx=Load()))])]))])"
        """
        tree = self.do_lambdef(st, ctx)
        tree.st = st
        return tree
    def do_or_test (self, st, ctx) :
        """or_test: and_test ('or' and_test)*
        -> ast.AST

        <<< x or y
        "Module(body=[Expr(value=BoolOp(op=Or(), values=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())]))])"
        <<< x or y or z
        "Module(body=[Expr(value=BoolOp(op=Or(), values=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load()), Name(id='z', ctx=Load())]))])"
        """
        return self._do_boolean(st, ctx)
    def do_and_test (self, st, ctx) :
        """and_test: not_test ('and' not_test)*
        -> ast.AST

        <<< x and y
        "Module(body=[Expr(value=BoolOp(op=And(), values=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())]))])"
        <<< x and y and z
        "Module(body=[Expr(value=BoolOp(op=And(), values=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load()), Name(id='z', ctx=Load())]))])"
        """
        return self._do_boolean(st, ctx)
    def do_not_test (self, st, ctx) :
        """not_test: 'not' not_test | comparison
        -> ast.AST

        <<< not x
        "Module(body=[Expr(value=UnaryOp(op=Not(), operand=Name(id='x', ctx=Load())))])"
        <<< not not x
        "Module(body=[Expr(value=UnaryOp(op=Not(), operand=UnaryOp(op=Not(), operand=Name(id='x', ctx=Load()))))])"
        """
        return self._do_unary(st, ctx)
    def do_comparison (self, st, ctx) :
        """comparison: star_expr (comp_op star_expr)*
        -> ast.AST

        <<< *x < *y <= *z
        "Module(body=[Expr(value=Compare(left=Starred(value=Name(id='x', ctx=Load()), ctx=Load()), ops=[Lt(), LtE()], comparators=[Starred(value=Name(id='y', ctx=Load()), ctx=Load()), Starred(value=Name(id='z', ctx=Load()), ctx=Load())]))])"
        <<< x < y <= z
        "Module(body=[Expr(value=Compare(left=Name(id='x', ctx=Load()), ops=[Lt(), LtE()], comparators=[Name(id='y', ctx=Load()), Name(id='z', ctx=Load())]))])"
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.Compare(lineno=st.srow, col_offset=st.scol,
                                   left=self.do(st[0], ctx),
                                   ops = [self.do(child, ctx)
                                          for child in st[1::2]],
                                   comparators = [self.do(child, ctx)
                                                  for child in st[2::2]])
    _comparison = {"<" : ast.Lt,
                   ">" : ast.Gt,
                   "==" : ast.Eq,
                   ">=" : ast.GtE,
                   "<=" : ast.LtE,
                   "!=" : ast.NotEq,
                   "<>" : ast.NotEq,
                   "in" : ast.In,
                   "not in" : ast.NotIn,
                   "is" : ast.Is,
                   "is not" : ast.IsNot}
    def do_comp_op (self, st, ctx) :
        """comp_op: '<'|'>'|'=='|'>='|'<='|'!='|'<>'|'in'|'not' 'in'|'is'
                |'is' 'not'
        -> ast.cmpop

        <<< 1 < 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[Lt()], comparators=[Num(n=2)]))])'
        <<< 1 > 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[Gt()], comparators=[Num(n=2)]))])'
        <<< 1 == 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[Eq()], comparators=[Num(n=2)]))])'
        <<< 1 >= 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[GtE()], comparators=[Num(n=2)]))])'
        <<< 1 <= 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[LtE()], comparators=[Num(n=2)]))])'
        <<< 1 != 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[NotEq()], comparators=[Num(n=2)]))])'
        <<< 1 <> 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[NotEq()], comparators=[Num(n=2)]))])'
        <<< 1 in 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[In()], comparators=[Num(n=2)]))])'
        <<< 1 not in 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[NotIn()], comparators=[Num(n=2)]))])'
        <<< 1 is 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[Is()], comparators=[Num(n=2)]))])'
        <<< 1 is not 2
        'Module(body=[Expr(value=Compare(left=Num(n=1), ops=[IsNot()], comparators=[Num(n=2)]))])'
        """
        text = " ".join(child.text for child in st)
        return self._comparison[text](lineno=st.srow, col_offset=st.scol)
    def do_star_expr (self, st, ctx) :
        """star_expr: ['*'] expr
        -> ast.AST

        <<< x
        "Module(body=[Expr(value=Name(id='x', ctx=Load()))])"
        <<< *x
        "Module(body=[Expr(value=Starred(value=Name(id='x', ctx=Load()), ctx=Load()))])"
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.Starred(lineno=st.srow, col_offset=st.scol,
                                   value=self.do(st[1], ctx),
                                   ctx=ctx())
    def do_expr (self, st, ctx) :
        """expr: xor_expr ('|' xor_expr)*
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1 | 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=BitOr(), right=Num(n=2)))])'
        <<< 1 | 2 | 3
        'Module(body=[Expr(value=BinOp(left=BinOp(left=Num(n=1), op=BitOr(), right=Num(n=2)), op=BitOr(), right=Num(n=3)))])'
        """
        return self._do_binary(st, ctx)
    def do_xor_expr (self, st, ctx) :
        """xor_expr: and_expr ('^' and_expr)*
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1 ^ 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=BitXor(), right=Num(n=2)))])'
        <<< 1 ^ 2 ^ 3
        'Module(body=[Expr(value=BinOp(left=BinOp(left=Num(n=1), op=BitXor(), right=Num(n=2)), op=BitXor(), right=Num(n=3)))])'
        """
        return self._do_binary(st, ctx)
    def do_and_expr (self, st, ctx) :
        """and_expr: shift_expr ('&' shift_expr)*
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1 & 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=BitAnd(), right=Num(n=2)))])'
        <<< 1 & 2 & 3
        'Module(body=[Expr(value=BinOp(left=BinOp(left=Num(n=1), op=BitAnd(), right=Num(n=2)), op=BitAnd(), right=Num(n=3)))])'
        """
        return self._do_binary(st, ctx)
    def do_shift_expr (self, st, ctx) :
        """shift_expr: arith_expr (('<<'|'>>') arith_expr)*
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1 << 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=LShift(), right=Num(n=2)))])'
        <<< 1 << 2 >> 3
        'Module(body=[Expr(value=BinOp(left=BinOp(left=Num(n=1), op=LShift(), right=Num(n=2)), op=RShift(), right=Num(n=3)))])'
        """
        return self._do_binary(st, ctx)
    def do_arith_expr (self, st, ctx) :
        """arith_expr: term (('+'|'-') term)*
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1 + 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=Add(), right=Num(n=2)))])'
        <<< 1 + 2 - 3
        'Module(body=[Expr(value=BinOp(left=BinOp(left=Num(n=1), op=Add(), right=Num(n=2)), op=Sub(), right=Num(n=3)))])'
        """
        return self._do_binary(st, ctx)
    def do_term (self, st, ctx) :
        """term: factor (('*'|'/'|'%'|'//') factor)*
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1 * 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=Mult(), right=Num(n=2)))])'
        <<< 1 * 2 / 3
        'Module(body=[Expr(value=BinOp(left=BinOp(left=Num(n=1), op=Mult(), right=Num(n=2)), op=Div(), right=Num(n=3)))])'
        <<< 1 * 2 / 3 % 4
        'Module(body=[Expr(value=BinOp(left=BinOp(left=BinOp(left=Num(n=1), op=Mult(), right=Num(n=2)), op=Div(), right=Num(n=3)), op=Mod(), right=Num(n=4)))])'
        <<< 1 * 2 / 3 % 4 // 5
        'Module(body=[Expr(value=BinOp(left=BinOp(left=BinOp(left=BinOp(left=Num(n=1), op=Mult(), right=Num(n=2)), op=Div(), right=Num(n=3)), op=Mod(), right=Num(n=4)), op=FloorDiv(), right=Num(n=5)))])'
        """
        return self._do_binary(st, ctx)
    def do_factor (self, st, ctx) :
        """factor: ('+'|'-'|'~') factor | power
        -> ast.AST

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< +1
        'Module(body=[Expr(value=UnaryOp(op=UAdd(), operand=Num(n=1)))])'
        <<< -1
        'Module(body=[Expr(value=Num(n=-1))])'
        <<< ~1
        'Module(body=[Expr(value=UnaryOp(op=Invert(), operand=Num(n=1)))])'
        <<< +-1
        'Module(body=[Expr(value=UnaryOp(op=UAdd(), operand=Num(n=-1)))])'
        <<< -+1
        'Module(body=[Expr(value=UnaryOp(op=USub(), operand=UnaryOp(op=UAdd(), operand=Num(n=1))))])'
        <<< +-~1
        'Module(body=[Expr(value=UnaryOp(op=UAdd(), operand=UnaryOp(op=USub(), operand=UnaryOp(op=Invert(), operand=Num(n=1)))))])'
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            tree = self._do_unary(st, ctx)
            if (isinstance(tree.op, self.ST.USub)
                and isinstance(tree.operand, self.ST.Num)
                and tree.operand.n > 0) :
                tree = self.ST.Num(lineno=st.srow, col_offset=st.scol,
                                   n = -tree.operand.n)
            return tree
    def do_power (self, st, ctx) :
        """power: atom trailer* ['**' factor]
        -> ast.AST

        <<< 1 ** 2
        'Module(body=[Expr(value=BinOp(left=Num(n=1), op=Pow(), right=Num(n=2)))])'
        <<< a.b ** 2
        "Module(body=[Expr(value=BinOp(left=Attribute(value=Name(id='a', ctx=Load()), attr='b', ctx=Load()), op=Pow(), right=Num(n=2)))])"
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            left = self.do(st[0], ctx)
            power = None
            for child in st[1:] :
                if child.text == "**" :
                    power = self.do(st[-1], ctx)
                    break
                trailer = self.do(child, ctx)
                left = trailer(left, st.srow, st.scol)
            if power :
                return self.ST.BinOp(lineno=st.srow, col_offset=st.scol,
                                     left=left,
                                     op=self.ST.Pow(lineno=st[-2].srow,
                                                    col_offset=st[-2].scol),
                                     right=power)
            else :
                return left
    def do_atom (self, st, ctx) :
        """atom: ('(' [yield_expr|testlist_comp] ')' |
            '[' [testlist_comp] ']' |
            '{' [dictorsetmaker] '}' |
            NAME | NUMBER | STRING+ | '...' | 'None' | 'True' | 'False')
        -> ast.AST

        <<< foo
        "Module(body=[Expr(value=Name(id='foo', ctx=Load()))])"
        <<< True
        "Module(body=[Expr(value=Name(id='True', ctx=Load()))])"
        <<< 42
        'Module(body=[Expr(value=Num(n=42))])'
        <<< 1.5
        'Module(body=[Expr(value=Num(n=1.5))])'
        <<< 'hello'
        "Module(body=[Expr(value=Str(s='hello'))])"
        <<< 'hello' 'world'
        "Module(body=[Expr(value=Str(s='helloworld'))])"
        <<< '''hello
        ... world'''
        "Module(body=[Expr(value=Str(s='hello\\\\nworld'))])"
        <<< [1, 2, 3]
        'Module(body=[Expr(value=List(elts=[Num(n=1), Num(n=2), Num(n=3)], ctx=Load()))])'
        <<< [x for x in l]
        "Module(body=[Expr(value=ListComp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        <<< (1, 2, 3)
        'Module(body=[Expr(value=Tuple(elts=[Num(n=1), Num(n=2), Num(n=3)], ctx=Load()))])'
        <<< (1,)
        'Module(body=[Expr(value=Tuple(elts=[Num(n=1)], ctx=Load()))])'
        <<< (1)
        'Module(body=[Expr(value=Num(n=1))])'
        <<< (x for x in l)
        "Module(body=[Expr(value=GeneratorExp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        <<< {1, 2, 3}
        'Module(body=[Expr(value=Set(elts=[Num(n=1), Num(n=2), Num(n=3)]))])'
        <<< {x for x in l}
        "Module(body=[Expr(value=SetComp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        <<< {1:2, 3:4}
        'Module(body=[Expr(value=Dict(keys=[Num(n=1), Num(n=3)], values=[Num(n=2), Num(n=4)]))])'
        <<< {x:y for x, y in l}
        "Module(body=[Expr(value=DictComp(key=Name(id='x', ctx=Load()), value=Name(id='y', ctx=Load()), generators=[comprehension(target=Tuple(elts=[Name(id='x', ctx=Store()), Name(id='y', ctx=Store())], ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        """
        kind, text = st[0].kind, st[0].text
        if kind == self.NUMBER :
            return self.ST.Num(lineno=st.srow, col_offset=st.scol,
                               n=self.ST.literal_eval(text))
        elif kind == self.NAME :
            return self.ST.Name(lineno=st.srow, col_offset=st.scol,
                                id=text, ctx=ctx())
        elif kind == self.STRING :
            return self.ST.Str(lineno=st.srow, col_offset=st.scol,
                               s="".join(self.ST.literal_eval(child.text)
                                         for child in st))
        elif text == "..." :
            return self.ST.Ellipsis(lineno=st.srow, col_offset=st.scol)
        elif text == "[" :
            if len(st) == 2 :
                return self.ST.List(lineno=st.srow, col_offset=st.scol,
                                    elts=[], ctx=ctx())
            else :
                loop, elts, atom = self.do(st[1], ctx)
                if atom is not None :
                    elts=[atom]
                if loop is None :
                    return self.ST.List(lineno=st.srow, col_offset=st.scol,
                                        elts=elts, ctx=ctx())
                else :
                    return self.ST.ListComp(lineno=st.srow, col_offset=st.scol,
                                            elt=loop, generators=elts)
        elif text == "(" :
            if len(st) == 2 :
                return self.ST.Tuple(lineno=st.srow, col_offset=st.scol,
                                     elts=[], ctx=ctx())
            elif st[1].symbol == "yield_expr" :
                return self.do(st[1], ctx)
            else :
                loop, elts, atom = self.do(st[1], ctx)
                if atom is not None :
                    return atom
                elif loop is None :
                    return self.ST.Tuple(lineno=st.srow, col_offset=st.scol,
                                         elts=elts, ctx=ctx())
                else :
                    return self.ST.GeneratorExp(lineno=st.srow, col_offset=st.scol,
                                                elt=loop, generators=elts)
        else : # text == "{"
            if len(st) == 2 :
                return self.ST.Dict(lineno=st.srow, col_offset=st.scol,
                                    keys=[], values=[])
            else :
                return self.do(st[1], ctx)
    def do_testlist_comp (self, st, ctx) :
        """testlist_comp: test ( comp_for | (',' test)* [','] )
        -> ast.AST?, ast.AST+

        <<< [1, 2, 3]
        'Module(body=[Expr(value=List(elts=[Num(n=1), Num(n=2), Num(n=3)], ctx=Load()))])'
        <<< [x for x in l]
        "Module(body=[Expr(value=ListComp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        """
        if len(st) == 1 :
            return None, None, self.do(st[0])
        elif st[1].text == "," :
            return None, [self.do(child, ctx) for child in st[::2]], None
        else :
            return self.do(st[0], ctx), self.do(st[1], ctx)[0], None
    def do_trailer (self, st, ctx) :
        """trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME
        -> (tree, line, column -> ast.AST)

        <<< a.b
        "Module(body=[Expr(value=Attribute(value=Name(id='a', ctx=Load()), attr='b', ctx=Load()))])"
        <<< a.b.c
        "Module(body=[Expr(value=Attribute(value=Attribute(value=Name(id='a', ctx=Load()), attr='b', ctx=Load()), attr='c', ctx=Load()))])"
        <<< a.b[c].d
        "Module(body=[Expr(value=Attribute(value=Subscript(value=Attribute(value=Name(id='a', ctx=Load()), attr='b', ctx=Load()), slice=Index(value=Name(id='c', ctx=Load())), ctx=Load()), attr='d', ctx=Load()))])"
        <<< f()
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=None))])"
        <<< f(x)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[], starargs=None, kwargs=None))])"
        <<< f(x, *l, y=2)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[keyword(arg='y', value=Num(n=2))], starargs=Name(id='l', ctx=Load()), kwargs=None))])"
        <<< f(x, *l, y=2, **d)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[keyword(arg='y', value=Num(n=2))], starargs=Name(id='l', ctx=Load()), kwargs=Name(id='d', ctx=Load())))])"
        <<< f(*l)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[], keywords=[], starargs=Name(id='l', ctx=Load()), kwargs=None))])"
        <<< f(**d)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=Name(id='d', ctx=Load())))])"
        <<< f(x)(a=1, b=2)
        "Module(body=[Expr(value=Call(func=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[], starargs=None, kwargs=None), args=[], keywords=[keyword(arg='a', value=Num(n=1)), keyword(arg='b', value=Num(n=2))], starargs=None, kwargs=None))])"
        """
        hd = st[0].text
        if hd == "." :
            def trail (tree, lineno, col_offset) :
                return self.ST.Attribute(lineno=lineno,
                                         col_offset=col_offset,
                                         value=tree,
                                         attr=st[1].text,
                                         ctx=ctx())
        elif hd == "[" :
            subscript = self.do(st[1], ctx)
            if len(subscript) == 1 :
                subscript = subscript[0]
            else :
                subscript = self.ST.ExtSlice(lineno=st[1].srow,
                                             col_offset=st[1].scol,
                                             dims=subscript,
                                             ctx=ctx())
            def trail (tree, lineno, col_offset) :
                return self.ST.Subscript(lineno=lineno,
                                         col_offset=col_offset,
                                         value=tree,
                                         slice=subscript,
                                         ctx=ctx())
        elif len(st) == 2 : # hd = "("
            def trail (tree, lineno, col_offset) :
                return self.ST.Call(lineno=lineno, col_offset=col_offset,
                                    func=tree, args=[], keywords=[],
                                    starargs=None, kwargs=None)
        else : # hd = "("
            def trail (tree, lineno, col_offset) :
                args, keywords, starargs, kwargs = self.do(st[1], ctx)
                return self.ST.Call(lineno=lineno, col_offset=col_offset,
                                    func=tree, args=args, keywords=keywords,
                                    starargs=starargs, kwargs=kwargs)
        return trail
    def do_subscriptlist (self, st, ctx) :
        """subscriptlist: subscript (',' subscript)* [',']
        -> ast.Slice+

        <<< l[:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=None), ctx=Load()))])"
        <<< l[1:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=None, step=None), ctx=Load()))])"
        <<< l[1::]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=None, step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[1:2:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=Num(n=2), step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[1:2:3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=Num(n=2), step=Num(n=3)), ctx=Load()))])"
        <<< l[::]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[:2:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=Num(n=2), step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[:2:3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=Num(n=2), step=Num(n=3)), ctx=Load()))])"
        <<< l[::3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=Num(n=3)), ctx=Load()))])"
        <<< l[1:2:3,4:5:6]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=ExtSlice(dims=[Slice(lower=Num(n=1), upper=Num(n=2), step=Num(n=3)), Slice(lower=Num(n=4), upper=Num(n=5), step=Num(n=6))]), ctx=Load()))])"
        """
        return [self.do(child, ctx) for child in st[::2]]
    def do_subscript (self, st, ctx) :
        """subscript: test | [test] ':' [test] [sliceop]
        -> ast.Slice | ast.Index

        <<< l[:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=None), ctx=Load()))])"
        <<< l[1:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=None, step=None), ctx=Load()))])"
        <<< l[1::]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=None, step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[1:2:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=Num(n=2), step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[1:2:3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=Num(n=2), step=Num(n=3)), ctx=Load()))])"
        <<< l[::]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[:2:]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=Num(n=2), step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[:2:3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=Num(n=2), step=Num(n=3)), ctx=Load()))])"
        <<< l[::3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=Num(n=3)), ctx=Load()))])"
        """
        count = len(st)
        if count == 1 and st[0].text == ":" :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=None, upper=None, step=None)
        elif count == 1 :
            return self.ST.Index(lineno=st.srow, col_offset=st.scol,
                                 value=self.do(st[0], ctx))
        elif count == 4 :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=self.do(st[0], ctx),
                                 upper=self.do(st[2], ctx),
                                 step=self.do(st[3], ctx))
        elif count == 3 and st[-1].symbol == "test" :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=self.do(st[0], ctx),
                                 upper=self.do(st[2], ctx),
                                 step=None)
        elif count == 3 and st[0].text == ":" :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=None,
                                 upper=self.do(st[1], ctx),
                                 step=self.do(st[2], ctx))
        elif count == 3 :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=self.do(st[0], ctx),
                                 upper=None,
                                 step=self.do(st[2], ctx))
        elif count == 2 and st[-1].symbol == "sliceop" :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=None,
                                 upper=None,
                                 step=self.do(st[1], ctx))
        elif count == 2 and st[0].text == ":" :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=None,
                                 upper=self.do(st[1], ctx),
                                 step=None)
        else :
            return self.ST.Slice(lineno=st.srow, col_offset=st.scol,
                                 lower=self.do(st[0], ctx),
                                 upper=None,
                                 step=None)
    def do_sliceop (self, st, ctx) :
        """sliceop: ':' [test]
        -> ast.AST

        <<< l[1::]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=Num(n=1), upper=None, step=Name(id='None', ctx=Load())), ctx=Load()))])"
        <<< l[::3]
        "Module(body=[Expr(value=Subscript(value=Name(id='l', ctx=Load()), slice=Slice(lower=None, upper=None, step=Num(n=3)), ctx=Load()))])"
        """
        if len(st) == 1 :
            return self.ST.Name(lineno=st.srow, col_offset=st.scol,
                                id="None", ctx=ctx())
        else :
            return self.do(st[1], ctx)
    def do_exprlist (self, st, ctx) :
        """exprlist: star_expr (',' star_expr)* [',']
        -> ast.AST+

        <<< del x
        "Module(body=[Delete(targets=[Name(id='x', ctx=Del())])])"
        <<< del x, y
        "Module(body=[Delete(targets=[Name(id='x', ctx=Del()), Name(id='y', ctx=Del())])])"
        """
        tree = self.do_testlist(st, ctx)
        tree.st = st
        return tree
    def do_testlist (self, st, ctx) :
        """testlist: test (',' test)* [',']
        -> ast.AST | ast.Tuple

        <<< 1
        'Module(body=[Expr(value=Num(n=1))])'
        <<< 1, 2
        'Module(body=[Expr(value=Tuple(elts=[Num(n=1), Num(n=2)], ctx=Load()))])'
        """
        lst = [self.do(child, ctx) for child in st[::2]]
        if len(lst) == 1 :
            return lst[0]
        else :
            return self.ST.Tuple(lineno=st.srow, col_offset=st.scol,
                                 elts=lst, ctx=ctx())
    def do_dictorsetmaker (self, st, ctx) :
        """dictorsetmaker: ( (test ':' test (comp_for | (',' test ':' test)* [','])) |
                  (test (comp_for | (',' test)* [','])) )
        -> ast.Dict | ast.DictComp | ast.Set | ast.SetComp

        <<< {1, 2, 3}
        'Module(body=[Expr(value=Set(elts=[Num(n=1), Num(n=2), Num(n=3)]))])'
        <<< {x for x in l}
        "Module(body=[Expr(value=SetComp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        <<< {1:2, 3:4}
        'Module(body=[Expr(value=Dict(keys=[Num(n=1), Num(n=3)], values=[Num(n=2), Num(n=4)]))])'
        <<< {x:y for x, y in l}
        "Module(body=[Expr(value=DictComp(key=Name(id='x', ctx=Load()), value=Name(id='y', ctx=Load()), generators=[comprehension(target=Tuple(elts=[Name(id='x', ctx=Store()), Name(id='y', ctx=Store())], ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])]))])"
        <<< {42}
        'Module(body=[Expr(value=Set(elts=[Num(n=42)]))])'
        """
        if len(st) == 1 :
            return self.ST.Set(lineno=st.srow, col_offset=st.scol,
                               elts=[self.do(st[0], ctx)])
        elif st[1].text == ":" :
            if len(st) < 4 or st[3].text == "," :
                return self.ST.Dict(lineno=st.srow, col_offset=st.scol,
                                    keys=[self.do(child, ctx)
                                          for child in st[::4]],
                                    values=[self.do(child, ctx)
                                            for child in st[2::4]])
            else :
                return self.ST.DictComp(lineno=st.srow, col_offset=st.scol,
                                        key=self.do(st[0], ctx),
                                        value=self.do(st[2], ctx),
                                        generators=self.do(st[3], ctx)[0])
        else :
            loop, elts, atom = self.do_testlist_comp(st, ctx)
            if loop is None :
                return self.ST.Set(lineno=st.srow, col_offset=st.scol,
                                   elts=elts)
            else :
                return self.ST.SetComp(lineno=st.srow, col_offset=st.scol,
                                       elt=loop, generators=elts)
    def do_classdef (self, st, ctx) :
        """classdef: 'class' NAME ['(' [arglist] ')'] ':' suite
        -> ast.ClassDef

        <<< class c : pass
        "Module(body=[ClassDef(name='c', bases=[], keywords=[], starargs=None, kwargs=None, body=[Pass()], decorator_list=[])])"
        <<< class c () : pass
        "Module(body=[ClassDef(name='c', bases=[], keywords=[], starargs=None, kwargs=None, body=[Pass()], decorator_list=[])])"
        <<< class c (object, foo=bar) : pass
        "Module(body=[ClassDef(name='c', bases=[Name(id='object', ctx=Load())], keywords=[keyword(arg='foo', value=Name(id='bar', ctx=Load()))], starargs=None, kwargs=None, body=[Pass()], decorator_list=[])])"
        """
        if len(st) <= 6 :
            return self.ST.ClassDef(lineno=st.srow, col_offset=st.scol,
                                    name=st[1].text,
                                    bases=[],
                                    keywords=[],
                                    starargs=None,
                                    kwargs=None,
                                    body=self.do(st[-1], ctx),
                                    decorator_list=[])
        else :
            args, keywords, starargs, kwargs = self.do(st[3], ctx)
            return self.ST.ClassDef(lineno=st.srow, col_offset=st.scol,
                                    name=st[1].text,
                                    bases=args,
                                    keywords=keywords,
                                    starargs=starargs,
                                    kwargs=kwargs,
                                    body=self.do(st[-1], ctx),
                                    decorator_list=[])
    def do_arglist (self, st, ctx) :
        """arglist: (argument ',')* (argument [',']
                         |'*' test (',' argument)* [',' '**' test]
                         |'**' test)
        -> args=ast.AST*, keywords=ast.keyword*, starargs=ast.AST?, kwargs=ast.AST?

        <<< f(x)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[], starargs=None, kwargs=None))])"
        <<< f(x, *l, y=2)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[keyword(arg='y', value=Num(n=2))], starargs=Name(id='l', ctx=Load()), kwargs=None))])"
        <<< f(x, *l, y=2, **d)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[keyword(arg='y', value=Num(n=2))], starargs=Name(id='l', ctx=Load()), kwargs=Name(id='d', ctx=Load())))])"
        <<< f(*l)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[], keywords=[], starargs=Name(id='l', ctx=Load()), kwargs=None))])"
        <<< f(**d)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[], keywords=[], starargs=None, kwargs=Name(id='d', ctx=Load())))])"
        <<< f(x=1, y=2, x=3)
        Traceback (most recent call last):
          ...
        ParseError: ... keyword argument repeated
        """
        args = []
        keywords = []
        allkw = set()
        starargs = None
        kwargs = None
        nodes = [n for n in st if n.text != ","]
        while nodes :
            if nodes[0].text == "*" :
                starargs = self.do(nodes[1], ctx)
                del nodes[:2]
            elif nodes[0].text == "**" :
                kwargs = self.do(nodes[1], ctx)
                del nodes[:2]
            else :
                arg = self.do(nodes[0], ctx)
                if isinstance(arg, self.ST.keyword) :
                    if arg.arg in allkw :
                        raise ParseError(nodes[0].text,
                                         reason="keyword argument repeated")
                    keywords.append(arg)
                    allkw.add(arg.arg)
                elif starargs is not None :
                    raise ParseError(nodes[0].text, reason="only named"
                                     " arguments may follow *expression")
                else :
                    args.append(arg)
                del nodes[0]
        return args, keywords, starargs, kwargs
    def do_argument (self, st, ctx) :
        """argument: test [comp_for] | test '=' test
        -> ast.keyword | ast.GeneratorExp

        <<< f(x)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[Name(id='x', ctx=Load())], keywords=[], starargs=None, kwargs=None))])"
        <<< f(x=1)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[], keywords=[keyword(arg='x', value=Num(n=1))], starargs=None, kwargs=None))])"
        <<< f(x for x in l)
        "Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load()), args=[GeneratorExp(elt=Name(id='x', ctx=Load()), generators=[comprehension(target=Name(id='x', ctx=Store()), iter=Name(id='l', ctx=Load()), ifs=[])])], keywords=[], starargs=None, kwargs=None))])"
        """
        test = self.do(st[0], ctx)
        if len(st) == 1 :
            return test
        elif len(st) == 3 :
            if not isinstance(test, self.ST.Name) :
                raise ParseError(st[0].text, reason="keyword can't be an"
                                 " expression")
            return self.ST.keyword(lineno=st.srow, col_offset=st.scol,
                                   arg=test.id,
                                   value=self.do(st[2], ctx))
        else :
            comp, ifs = self.do(st[1], ctx)
            return self.ST.GeneratorExp(lineno=st.srow, col_offset=st.scol,
                                        elt=test, generators=comp)
    def do_comp_iter (self, st, ctx) :
        """comp_iter: comp_for | comp_if
        -> comprehension*, ast.AST*

        <<< [a for b in c if d if e for f in g if h]
        "Module(body=[Expr(value=ListComp(elt=Name(id='a', ctx=Load()), generators=[comprehension(target=Name(id='b', ctx=Store()), iter=Name(id='c', ctx=Load()), ifs=[Name(id='d', ctx=Load()), Name(id='e', ctx=Load())]), comprehension(target=Name(id='f', ctx=Store()), iter=Name(id='g', ctx=Load()), ifs=[Name(id='h', ctx=Load())])]))])"
        """
        return self.do(st[0], ctx)
    def do_comp_for (self, st, ctx) :
        """comp_for: 'for' exprlist 'in' or_test [comp_iter]
        -> comprehension+, []

        <<< [a for b in c if d if e for f in g if h]
        "Module(body=[Expr(value=ListComp(elt=Name(id='a', ctx=Load()), generators=[comprehension(target=Name(id='b', ctx=Store()), iter=Name(id='c', ctx=Load()), ifs=[Name(id='d', ctx=Load()), Name(id='e', ctx=Load())]), comprehension(target=Name(id='f', ctx=Store()), iter=Name(id='g', ctx=Load()), ifs=[Name(id='h', ctx=Load())])]))])"
        """
        if len(st) == 4 :
            return [self.ST.comprehension(lineno=st.srow,
                                          col_offset=st.scol,
                                          target=self.do(st[1], ast.Store),
                                          iter=self.do(st[3], ctx),
                                          ifs=[])], []
        else :
            comp, ifs = self.do(st[4], ctx)
            return [self.ST.comprehension(lineno=st.srow,
                                          col_offset=st.scol,
                                          target=self.do(st[1], ast.Store),
                                          iter=self.do(st[3], ctx),
                                          ifs=ifs)] + comp, []
    def do_comp_if (self, st, ctx) :
        """comp_if: 'if' test_nocond [comp_iter]
        -> comprehension*, ast.AST+

        <<< [a for b in c if d if e for f in g if h]
        "Module(body=[Expr(value=ListComp(elt=Name(id='a', ctx=Load()), generators=[comprehension(target=Name(id='b', ctx=Store()), iter=Name(id='c', ctx=Load()), ifs=[Name(id='d', ctx=Load()), Name(id='e', ctx=Load())]), comprehension(target=Name(id='f', ctx=Store()), iter=Name(id='g', ctx=Load()), ifs=[Name(id='h', ctx=Load())])]))])"
        """
        if len(st) == 2 :
            return [], [self.do(st[1], ctx)]
        else :
            comp, ifs = self.do(st[2], ctx)
            return comp, [self.do(st[1], ctx)] + ifs
    def do_yield_expr (self, st, ctx) :
        """yield_expr: 'yield' [testlist]
        -> ast.Yield

        <<< yield
        'Module(body=[Expr(value=Yield(value=None))])'
        <<< yield 42
        'Module(body=[Expr(value=Yield(value=Num(n=42)))])'
        <<< yield 42, 43
        'Module(body=[Expr(value=Yield(value=Tuple(elts=[Num(n=42), Num(n=43)], ctx=Load())))])'
        """
        if len(st) == 2 :
            return self.ST.Yield(lineno=st.srow, col_offset=st.scol,
                                 value=self.do(st[1], ctx))
        else :
            return self.ST.Yield(lineno=st.srow, col_offset=st.scol,
                                 value=None)
    @classmethod
    def parse (cls, expr, mode="exec", filename="<string>") :
        tree = cls(cls.parser.parseString(expr.strip() + "\n",
                                          filename=filename)).ast
        if mode == "exec" :
            return tree
        elif mode == "eval" :
            if len(tree.body) > 1 or not isinstance(tree.body[0], cls.ST.Expr) :
                raise ParseError(None, reason="invalid syntax")
            return cls.ST.Expression(body=tree.body[0].value)
        elif mode == "single" :
            return cls.ST.Interactive(body=tree.body)
        else :
            raise ValueError("arg 2 must be 'exec', 'eval' or 'single'")

parse = Translator.parse

class ParseTestParser (doctest.DocTestParser) :
    _EXAMPLE_RE = re.compile(r'''
        # Source consists of a PS1 line followed by zero or more PS2 lines.
        (?P<source>
            (?:^(?P<indent> [ ]*) <<<    .*)    # PS1 line
            (?:\n           [ ]*  \.\.\. .*)*)  # PS2 lines
        \n?
        # Want consists of any non-blank lines that do not start with PS1.
        (?P<want> (?:(?![ ]*$)    # Not a blank line
                     (?![ ]*<<<)  # Not a line starting with PS1
                     .*$\n?       # But any other line
                  )*)
        ''', re.MULTILINE | re.VERBOSE)
    def __init__ (self, translator) :
        self.Translator = translator
    def parse (self, string, name="<string>") :
        examples = doctest.DocTestParser.parse(self, string, name)
        try :
            rule = name.split(".do_", 1)[1]
        except :
            rule = None
        for i, exple in enumerate(examples) :
            if isinstance(exple, str) :
                continue
            elif name.split(".")[1] not in self.Translator.__dict__ :
                examples[i] = ("skipping example for another language: %r"
                               % exple.source)
                continue
            if rule is not None :
                source = exple.source.strip() + "\n"
                try :
                    tree = self.Translator.ParseTree(self.Translator.parser.parseString(source))
                except :
                    print("could not parse %r at %s:%s" % (source, name,
                                                           exple.lineno))
                    raise
                if not tree.involve(self.Translator.parser.stringMap[rule]) :
                    print(("test at %s:%s does not involve rule %s"
                           % (name, exple.lineno, rule)))
                    examples[i] = "<test skipped>"
                    continue
            examples[i] = doctest.Example(
                source=("ast.dump(parse(%r))") % exple.source,
                want=exple.want,
                exc_msg=exple.exc_msg,
                lineno=exple.lineno,
                indent=exple.indent,
                options=exple.options)
        return examples

def testparser (translator) :
    for rule in translator.parser.stringMap :
        try :
            assert "<<<" in getattr(translator, "do_" + rule).__doc__
        except AttributeError :
            print("no handler for rule %r" % rule)
            continue
        except TypeError :
            print("missing doc for rule %r" % rule)
            continue
        except AssertionError :
            print("missing test for rule %r" % rule)
            continue
    finder = doctest.DocTestFinder(parser=ParseTestParser(translator))
    runner = doctest.DocTestRunner(optionflags=doctest.NORMALIZE_WHITESPACE
                                   | doctest.ELLIPSIS)
    for name, method in inspect.getmembers(translator, inspect.ismethod) :
        if not name.startswith("do_") :
            continue
        for test in finder.find(method, "%s.%s" % (translator.__name__, name)) :
            runner.run(test)
    runner.summarize()

if __name__ == "__main__" :
    testparser(Translator)



import heapq, time
from snakes.nets import StateGraph
import snakes.lang
import snkast as ast

class Checker (object) :
    def __init__ (self, net, progress) :
        self.p = progress
        self.g = StateGraph(net)
        self.f = [self.build(f) for f in net.label("asserts")]
        if progress :
            if len(self.f) == 0 :
                print("WARNING: no assertion given (computing the state space anyway)")
            elif len(self.f) == 1 :
                print("checking 1 assertion")
            else :
                print("checking %s assertions" % len(self.f))
    def build (self, tree) :
        src = """
def check (_) :
    return %s
""" % tree.st.source()[7:]
        ctx = dict(self.g.net.globals)
        ctx["bounded"] = self.bounded
        ctx["dead"] = self.dead
        exec(src, ctx)
        fun = ctx["check"]
        fun.lineno = tree.lineno
        return fun
    def bounded (self, marking, max) :
        return all(len(marking(p)) == 1 for p in marking)
    def dead (self) :
        for _ in self.g.successors() :
            return False
        return True
    def run (self) :
        start = last = time.time()
        for state in self.g :
            marking = self.g.net.get_marking()
            for place in marking :
                if max(marking(place).values()) > 1 :
                    return len(self.g), None, self.trace(state)
            for check in self.f :
                try :
                    if not check(marking) :
                        return len(self.g), check.lineno, self.trace(state)
                except :
                    pass
            if self.p and time.time() - last >= 5 :
                last = time.time()
                print(" ... %s states explored so far in %.0f seconds"
                      % (len(self.g), last - start))
        return len(self.g), None, None
    def path (self, tgt, src=0) :
        q = [(0, src, ())]
        visited = set()
        while True :
            (c, v1, path) = heapq.heappop(q)
            if v1 not in visited :
                path = path + (v1,)
                if v1 == tgt :
                    return path
                visited.add(v1)
                for v2 in self.g.successors(v1) :
                    if v2 not in visited :
                        heapq.heappush(q, (c+1, v2, path))
    def trace (self, state) :
        path = self.path(state)
        return tuple(self.g.successors(i)[j]
                     for i, j in zip(path[:-1], path[1:]))



import snakes.plugins
snakes.plugins.load("status", "snakes.nets", "nets")
from nets import *
from dolev_yao import Nonce

ns = loads(",ns.pnml")
states = StateGraph(ns)

for s in states :
    m = states.net.get_marking()
    # skip non final markings
    if "bob.x" not in m or "alice.x" not in m :
        continue
    # get Alice's and Bob's peers ids
    bp = list(m["bob.peer"])[0]
    ap = list(m["alice.peer"])[0]
    # violation of mutual authentication
    if bp == 1 and ap != 2 :
        print(s, "A(1) <=> %s ; B(2) <=> %s" % (ap, bp))
        print(m)

print(len(states), "states")



from snakes.lang.ctlstar.parser import parse
from snakes.lang.ctlstar import asdl as ast
from snakes.lang import bind
import _ast

class SpecError (Exception) :
    def __init__ (self, node, reason) :
        Exception.__init__(self, "[line %s] %s" % (node.lineno, reason))

def astcopy (node) :
    if not isinstance(node, _ast.AST) :
        return node
    attr = {}
    for name in node._fields + node._attributes :
        value = getattr(node, name)
        if isinstance(value, list) :
            attr[name] = [astcopy(child) for child in value]
        else :
            attr[name] = astcopy(value)
    return node.__class__(**attr)

class Builder (object) :
    def __init__ (self, spec) :
        self.spec = spec
        self.decl = {}
        for node in spec.atoms + spec.properties :
            if node.name in self.decl :
                raise SpecError(node, "%r already declare line %s"
                                % (name.name, self.decl[name.name].lineno))
            self.decl[node.name] = node
        self.main = spec.main
    def build (self, node) :
        node = astcopy(node)
        return self._build(node, {})
    def _build (self, node, ctx) :
        if isinstance(node, ast.atom) :
            try :
                builder = getattr(self, "_build_%s" % node.__class__.__name__)
            except AttributeError :
                node.atomic = True
            else :
                node = builder(node, ctx)
                node.atomic = True
        elif isinstance(node, ast.CtlBinary) :
            node.left = self._build(node.left, ctx)
            node.right = self._build(node.right, ctx)
            node.atomic = (isinstance(node.op, (ast.boolop, ast.Imply,
                                                ast.Iff))
                           and node.left.atomic
                           and node.right.atomic)
        elif isinstance(node, ast.CtlUnary) :
            node.child = self._build(node.child, ctx)
            node.atomic = (isinstance(node.op, ast.Not)
                           and node.child.atomic)
        else :
            raise SpecError(node, "not a CTL* formula")
        return node
    def _build_place (self, param, ctx) :
        if isinstance(param, ast.Parameter) :
            if param.name not in ctx :
                raise SpecError(param, "place %r should be instantiated"
                                % param.name)
            return ctx[param.name]
        else :
            return param
    def _build_InPlace (self, node, ctx) :
        node.data = [bind(child, ctx) for child in node.data]
        node.place = self._build_place(node.place, ctx)
        return node
    def _build_NotInPlace (self, node, ctx) :
        return self._build_InPlace(node, ctx)
    def _build_EmptyPlace (self, node, ctx) :
        node.place = self._build_place(node.place, ctx)
        return node
    def _build_MarkedPlace (self, node, ctx) :
        return self._build_EmptyPlace(node, ctx)
    # skip Deadlock and Boolean: nothing to do
    def _build_Quantifier (self, node, ctx) :
        node.place = self._build_place(node.place, ctx)
        ctx = ctx.copy()
        for name in node.vars :
            ctx[name] = ast.Token(name, node.place.place)
        node.child = self._build(node.child, ctx)
        return node
    def _build_Instance (self, node, ctx) :
        if node.name not in self.decl :
            raise SpecError(node, "undeclared object %r" % node.name)
        ctx = ctx.copy()
        decl = self.decl[node.name]
        for arg in decl.args :
            ctx[arg.name] = arg
        if isinstance(decl, ast.Property) :
            return self._build_Instance_Property(node, decl, ctx)
        else :
            return self._build_Instance_Atom(node, decl, ctx)
    def _build_Instance_Property (self, node, prop, ctx) :
        bound = set(a.name for a in prop.args)
        args = dict((a.arg, a.annotation) for a in node.args)
        for param in prop.params :
            if param.name in bound :
                raise SpecError(node, "argument %r already bound"
                                % param.name)
            elif param.name in args :
                arg = args.pop(param.name)
                bound.add(param.name)
            else :
                raise SpecError(node, "missing argument %r" % param.name)
            if param.type == "place" :
                if not isinstance(arg, ast.Place) :
                    raise SpecError(node, "expected place for %r"
                                    % param.name)
                arg.name = param.name
            ctx[param.name] = arg
        if args :
            raise SpecError(node, "too many arguments (%s)"
                            % ", ".join(repr(a) for a in args))
        return self._build(astcopy(prop.body), ctx)
    def _build_Instance_Atom (self, node, atom, ctx) :
        bound = set(a.name for a in atom.args)
        args = dict((a.arg, a.annotation) for a in node.args)
        new = astcopy(atom)
        for param in atom.params :
            if param.name in bound :
                raise SpecError(node, "argument %r already bound"
                                % param.name)
            elif param.name in args :
                arg = args.pop(param.name)
                bound.add(param.name)
            else :
                raise SpecError(node, "missing argument %r" % param.name)
            if param.type == "place" :
                if not isinstance(arg, ast.Place) :
                    raise SpecError(node, "expected place for %r"
                                    % param.name)
                arg.name = param.name
            else :
                arg = ast.Argument(name=param.name,
                                   value=arg,
                                   type=param.type)
            new.args.append(arg)
        if args :
            raise SpecError(node, "too many arguments (%s)"
                            % ", ".join(repr(a) for a in args))
        del new.params[:]
        return new

def build (spec) :
    if isinstance(spec, str) :
        spec = parse(spec)
    return Builder(spec).build(spec.main)



import json as JSON

def _tag (name) :
    return name.lower().strip("_")

class Tag (object) :
    default = {"a" : {"href" : "#"}}
    noclose = set([None, "br"])
    def __init__ (self, name, *children, **attr) :
        if name is not None :
            self.name = name.lower()
        else :
            self.name = None
        self.attr = dict(self._cleanup(attr))
        self.children = list(children)
    def _cleanup (self, attr) :
        for key, value in attr.items() :
            yield _tag(key), value
    def __call__ (self, *children, **attr) :
        self.children.extend(children)
        self.attr.update(self._cleanup(attr))
        return self
    def add (self, *children) :
        self.children.extend(children)
        return self
    def join (self, children) :
        if self.name not in self.noclose :
            raise ValueError("cannot join with tag %r" % self.name)
        lst = [children[0]]
        for c in children[1:] :
            lst.extend([self, c])
        return self.__class__(None, *lst)
    def __setitem__ (self, key, value) :
        self(**{_tag(key): value})
    def __delitem__ (self, key) :
        self.attr.pop(_tag(key), None)
    def __contains__ (self, key) :
        return _tag(key) in self.attr
    def __str__ (self) :
        attr = self.attr.copy()
        for key, value in self.default.get(self.name, {}).items() :
            if key not in attr :
                attr[key] = value
        if self.name is None :
            return "".join(str(c) for c in self.children)
        elif attr :
            ret = "<%s %s>%s" % (
                self.name,
                " ".join("%s=%r" % a for a in attr.items()),
                "".join(str(c) for c in self.children))
        else :
            ret = "<%s>%s" % (
                self.name,
                "".join(str(c) for c in self.children))
        if self.children or self.name not in self.noclose :
            return ret + "</%s>" % self.name
        else :
            return ret[:-1] + "/>"
    def __repr__ (self) :
        return repr(str(self))

class Factory (object) :
    def __getattr__ (self, name) :
        return Tag(_tag(name))

H = Factory()

def utf8 (text) :
    return text.encode("utf-8")

class JSONEncoder(JSON.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Tag):
            return utf8(str(obj))
        elif isinstance(obj, (unicode, str)) :
            return utf8(obj)
        else :
            return JSON.JSONEncoder.default(self, obj)

json = JSONEncoder().encode



from snakes.compat import io
from snakes.lang.abcd.parser import ast, parse
import os, tempfile, re, codecs, collections
try :
    from cgi import escape
except :
    from html import escape

template_css = u"""/* ABCD source code */
  .abcd { border:solid 1px #DDD; border-radius:5px; padding:5px 10px; margin:5px; background-color:#F4F4F4; overflow:auto; }
  .abcd .comment { color:#888; }
  .abcd .ident { color:#808; }
  .abcd .string { color:#088; }
  .abcd .kw { color:#800; font-weight:bold; }
  .abcd .flow { color:#800; font-weight:bold; }
  .abcd .buffer .decl { color:#080; font-weight:bold; }
  .abcd .net .decl { color:#008; font-weight:bold; }
  .abcd .instance .name { color:#008; }
  .abcd .action .delim { font-weight:bold; }
  .abcd .action .name { color:#080; }
  .abcd .highlight { background-color:yellow; }
/* Petri net picture */
  .petrinet { border:solid 1px #DDD; border-radius:5px; padding:5px 10px; margin:5px; background-color:#FFF; overflow:auto; clear:both; }
/* Objects tree */
  .tree { border:solid 1px #DDD; border-radius:5px; padding:5px 10px; margin:5px; background-color:#F4F4F4; overflow:auto; font-family:monospace; }
  .tree .kw { color:#800; font-weight:bold; }
  .tree .buffer { color:#080; font-weight:bold; }
  .tree .ident { color:#808; }
  .tree .instance .name { color:#008; }
  .tree .action .delim { font-weight:bold; }
  .tree .action .name { color:#080; }
  .tree .string { color:#088; }
  .tree .highlight { background-color:yellow; }
"""

template_html = u'''<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head><title>%(filename)s</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>
<style type="text/css">%(css)s</style>
%(headers)s
%(jscode)s
</head><body>
<h1><tt>%(filename)s</tt></h1>
%(abcd)s
%(tree)s
%(svg)s
</body></html>
'''

template_tree = '''<div class="tree">%(tree)s</div>'''

template_headers = '''<script src="http://code.jquery.com/jquery-1.11.0.min.js" type="text/javascript"></script>'''

template_jscode = '''<script type="text/javascript">
var nodeColor;
function abcdon () {
  obj = jQuery(this);
  if (obj.attr("class") == "node") {
    node = obj.children().children().first();
    nodeColor = node.attr("fill");
    node.attr("fill", "yellow");
  } else {
    obj.addClass("highlight");
  }
  jQuery(obj.attr("data-abcd")).addClass("highlight");
};
function abcdoff () {
  obj = jQuery(this);
  if (obj.attr("class") == "node") {
    node = obj.children().children().first();
    node.attr("fill", nodeColor);
  } else {
    obj.removeClass("highlight");
  }
  jQuery(obj.attr("data-abcd")).removeClass("highlight");
};
function treeon () {
  obj = jQuery(this);
  if (obj.attr("class") != "node") {
    obj.addClass("highlight");
  }
  jQuery(obj.attr("data-tree")).addClass("highlight");
};
function treeoff () {
  obj = jQuery(this);
  if (obj.attr("class") != "node") {
    obj.removeClass("highlight");
  }
  jQuery(obj.attr("data-tree")).removeClass("highlight");
};
function neton () {
  obj = jQuery(this);
  jQuery(obj.attr("data-net")).each(function () {
      node = jQuery(this).children().children().first();
      nodeColor = node.attr("fill");
      node.attr("fill", "yellow");
  });
  obj.addClass("highlight");
};
function netoff () {
  obj = jQuery(this);
  jQuery(obj.attr("data-net")).each(function () {
      node = jQuery(this).children().children().first();
      node.attr("fill", nodeColor);
  });
  obj.removeClass("highlight");
};
function setwidth () {
  abcd = jQuery(".abcd");
  tree = jQuery(".tree");
  width = (jQuery(".petrinet").outerWidth(true) / 2) - 35;
  console.log(width);
  height = Math.max(abcd.height(), tree.height());
  abcd.css({float: "left", width: width, height: height});
  tree.css({float: "right", width: width, height: height});
}
jQuery(document).ready(function() {
  jQuery("[data-abcd]").hover(abcdon, abcdoff);
  jQuery("[data-tree]").hover(treeon, treeoff);
  jQuery("[data-net]").hover(neton, netoff);
  jQuery(".tree .instance, .tree .action").each(function () {
    obj = jQuery(this);
    obj.html(jQuery(obj.attr("data-abcd")).html());
  });
  setwidth();
  jQuery(window).resize(setwidth);
});
</script>'''

class Span (object) :
    def __init__ (self, cls=[], id=None, net=[], abcd=[], tree=[]) :
        self.cls = set([cls]) if isinstance(cls, str) else set(cls)
        self.id = id
        self.net = set([net]) if isinstance(net, str) else set(net)
        self.abcd = set([abcd]) if isinstance(abcd, str) else set(abcd)
        self.tree = set([tree]) if isinstance(tree, str) else set(tree)
    def copy (self, **attr) :
        span = self.__class__(cls=self.cls, id=self.id, net=self.net,
                              abcd=self.abcd, tree=self.tree)
        for k, v in attr.items() :
            if isinstance(getattr(span, k), set) :
                v = set(v)
            setattr(span, k, v)
        return span
    def __bool__ (self) :
        return self.__nonzero__()
    def __nonzero__ (self) :
        return bool(self.cls or self.id or self.net or self.abcd
                    or self.tree)
    def span (self) :
        attr = []
        for src, dst in (("cls", "class"), ("id", "id"),
                         ("net", "data-net"), ("abcd", "data-abcd"),
                         ("tree", "data-tree")) :
            val = getattr(self, src)
            if not val :
                continue
            elif isinstance(val, str) :
                attr.append("%s=%r" % (dst, val))
            elif dst.startswith("data-") :
                attr.append("%s=%r" % (dst, ", ".join("#" + v for v in val)))
            else :
                attr.append("%s=%r" % (dst, ", ".join(val)))
        return "<span %s>" % " ".join(attr)
    def __str__ (self) :
        return self.span()

keywords = {"False", "True", "and", "as", "assert", "buffer", "const",
            "else", "enum", "for", "from", "if", "import", "in", "is",
            "lambda", "net", "not", "or", "symbol", "task", "typedef"}

class ABCD2HTML (ast.NodeVisitor) :
    def __init__ (self, tree) :
        self.tree = tree
        self.x = "%%0%uX" % len("%x" % (tree.st.erow + 1))
        self.st = {}
        self.path = []
        self.nets = {}
        self.visit(tree)
    def __getitem__ (self, id) :
        cls = id.__class__.__name__
        if cls in ("Place", "Transition") :
            x, y, _, _ = id.label("srcloc")
            c = "B" if cls == "Place" else "A"
            id = self.newid(c, x, y)
        elif isinstance(id, tuple) :
            id = self.newid(*id)
        if isinstance(self.st[id], Span) :
            return self.st[id]
        else :
            return self.st[id].span
    def newid (self, c, x, y) :
        return "".join([c[0].upper(), self.x % x, "%X" % y])
    def setspan (self, cls, node, **args) :
        if isinstance(node, ast.AST) :
            node = node.st
        if cls in ("action", "buffer", "instance", "net") :
            x, y = node.srow, node.scol
            ident = self.newid(cls, x, y)
            ident = args.pop("id", ident)
            span = node.span = Span(cls=cls.lower(), id=ident, **args)
            self.st[ident] = node
        elif node is None :
            span = Span(cls=cls.lower(), **args)
            if "id" in args :
                self.st[args["id"]] = span
        else :
            span = node.span = Span(cls=cls.lower(), **args)
            if "id" in args :
                self.st[args["id"]] = node
        return span
    def visit_AbcdBuffer (self, node) :
        t = node.st
        while t.symbol != "abcd_buffer" :
            t = t[0]
        self.setspan("decl", t[1])
        self.setspan("buffer", node)
        self.generic_visit(node)
    def visit_AbcdNet (self, node) :
        t = node.st
        while t.symbol != "abcd_net" :
            t = t[0]
        self.setspan("decl", t[1])
        self.setspan("body", node.body)
        span = self.setspan("net", node)
        self.setspan("proto", None, id="P" + span.id)
        self.nets[".".join(self.path + [node.name])] = span.id
        self.path.append(node.name)
        self.generic_visit(node)
        self.path.pop(-1)
    def visit_AbcdInstance (self, node) :
        t = node.st
        while t.symbol != "abcd_instance" :
            t = t[0]
        self.setspan("name", t[0])
        pid = "P" + self.nets[".".join(self.path + [node.net])]
        span = self.setspan("instance", node, abcd=[pid])
        self[pid].abcd.add(span.id)
        self.generic_visit(node)
    def visit_AbcdAction (self, node) :
        t = node.st
        while t[0].text != '[' :
            t = t[0]
        span = self.setspan("action", node)
        self.setspan("delim", t[0], id="L" + span.id)
        self.setspan("delim", t[-1], id="R" + span.id)
        self.generic_visit(node)
    def visit_AbcdFlowOp (self, node) :
        t = node.st
        while len(t) == 1 :
            t = t[0]
            while t[0].text == '(' :
                t = t[1]
        for op in t[1::2] :
            self.setspan("flow", op)
        self.generic_visit(node)
    def visit_SimpleAccess (self, node) :
        self.setspan("name", node.st[0])
        self.setspan("access", node)
        self.generic_visit(node)
    def visit_FlushAccess (self, node) :
        self.visit_SimpleAccess(node)
    def visit_SwapAccess (self, node) :
        self.visit_SimpleAccess(node)
    def escape (self, text) :
        return escape(text)
    def build (self, st) :
        # collect text skipped during parsing (blanks and comments)
        output = io.StringIO()
        if st.srow > self.row :
            for line in st.text.lexer.lines[self.row-1:st.srow-1] :
                output.write(line[self.col:])
                self.col = 0
            output.write(st.text.lexer.lines[st.srow-1][:st.scol])
        elif st.scol > self.col :
            output.write(st.text.lexer.lines[self.row-1][self.col:st.scol])
        # insert skipped text with comments rendering
        for line in output.getvalue().splitlines(True) :
            if "#" in line :
                left, right = line.split("#", 1)
                self.output.write("%s<span class=%r>%s</span>"
                                  % (self.escape(left),
                                     "comment",
                                     self.escape("#" + right)))
            else :
                self.output.write(self.escape(line))
        # adjust current position in source code
        self.row, self.col = st.srow, st.scol
        # close span for net declaration
        span = getattr(st, "span", Span())
        if "body" in (c.lower() for c in span.cls) :
            self.output.write("</span>")
        # generate <span ...> if necessary
        if span :
            self.output.write(str(span))
        # generate span for net declaration
        if (span.id or "").startswith("N") :
            self.output.write(str(self["P" + span.id]))
        # render tree or its children
        if len(st) :
            for child in st :
                # add span tags on special elements
                if not hasattr(child, "span") :
                    if child.symbol == "NAME" :
                        if child.text in keywords :
                            self.setspan("kw", child)
                        else :
                            self.setspan("ident", child)
                    elif child.symbol == "STRING" :
                        self.setspan("string", child)
                    elif child.symbol == "COLON" :
                        self.setspan("kw", child)
                self.build(child)
        else :
            if st.symbol not in ("DEDENT", "ENDMARKER") :
                self.output.write(self.escape(st.text))
            self.row, self.col = st.erow, st.ecol
        # generate </span> if necessary
        if span :
            self.output.write("</span>")
    def html (self) :
        self.output = io.StringIO()
        self.indent, self.row, self.col = False, 1, 0
        self.build(self.tree.st)
        return "<pre class='abcd'>%s</pre>" % self.output.getvalue()

def Tree () :
    return collections.defaultdict(Tree)

class TreeInfo (object) :
    def __init__ (self, span, name) :
        self.span, self.name = span, name
    def __hash__ (self) :
        return hash(self.name)
    def __eq__ (self, other) :
        try :
            return self.name == other.name
        except :
            return False
    def __ne__ (self, other) :
        return not self.__eq__(other)
    def __iter__ (self) :
        yield self.span
        yield self.name

_svgclean = [(re.compile(r, re.I), s) for r, s in
             [(r"<[?!][^>]*>\n*", ""),
              (r"<title>[^<>]*</title>\n*", ""),
              (r"<g [^<>]*></g>\n*", ""),
              ]]

class Net2HTML (object) :
    def __init__ (self, net, gv, abcd) :
        self.gv = gv
        self.abcd = abcd
        self.tree = Tree()
        self.n2a = collections.defaultdict(set)
        self.n2t = {}
        snk = net.label("snakes")
        self.count = collections.defaultdict(int)
        for place in net.place() :
            nid = gv.nodemap[place.name]
            if place.status in (snk.entry, snk.internal, snk.exit) :
                for char, trans in ([("R", net.transition(t))
                                     for t in place.pre]
                                    + [("L", net.transition(t))
                                       for t in place.post]) :
                    span = abcd[trans]
                    self.n2a[nid].add(char + span.id)
            else :
                self.addtree(0, "buffer", place)
        for trans in net.transition() :
            self.addtree(10, "action", trans)
    def addtree (self, weight, kind, node) :
        nid = self.gv.nodemap[node.name]
        aid = self.abcd[node]
        tid = aid.copy(id=("T%X" % self.count[aid.id]) + aid.id,
                       tree=[], abcd=[aid.id], net=[nid])
        self.count[aid.id] += 1
        aid.tree.add(tid.id)
        aid.net.add(nid)
        self.n2a[nid].add(aid.id)
        self.n2t[nid] = tid.id
        pos = self.tree
        path = node.label("path")
        try :
            inst = node.label("instances")
        except :
            inst = [None] * len(path)
        for name, (_, srow, scol, _, _) in zip(path, inst) :
            a = self.abcd["I", srow, scol]
            t = a.copy(id=("T%X" % self.count[a.id]) + a.id,
                       tree=[], abcd=[a.id], net=[])
            self.count[a.id] += 1
            a.tree.add(t.id)
            pos = pos[((20, srow, scol), "instance", TreeInfo(t, name))]
        prefix = sum(len(p) for p in path) + len(path)
        srow, scol, _, _ = node.label("srcloc")
        pos[((weight, srow, scol), kind, (node, node.name[prefix:]))] = tid
    def _tree (self, tree, indent="") :
        yield indent + "<ul>"
        for (_, kind, data), child in sorted(tree.items()) :
            if kind == "instance" :
                yield indent + "<li>%s%s</span>" % tuple(data)
                for item in self._tree(child, indent + "  ") :
                    yield item
                yield indent + "</li>"
            else :
                node, name = data
                if kind == "buffer" :
                    content = ("<span class='kw'>buffer</span> "
                               "<span class='name'>%s</span> = "
                               "<span class='content'>%s</span>"
                               % (name, node.tokens))
                    yield indent + "<li>%s%s</span></li>" % (child, content)
                elif kind == "action" :
                    content = name
                    yield (indent + "<li>%s%s</span><ul class='modes'>"
                           + "</ul></li>") % (child, content)
                else :
                    raise ValueError("unexpected data %r" % kind)
        yield indent + "</ul>"
    def html (self) :
        return template_tree % {"tree" : "\n".join(self._tree(self.tree))}
    def svg (self) :
        # load SVG file
        with tempfile.NamedTemporaryFile(suffix=".svg") as tmp :
            self.gv.render(tmp.name)
            with codecs.open(tmp.name, "r", "utf-8") as infile :
                svg = infile.read()
        for r, s in _svgclean :
            svg = r.sub(s, svg)
        for node, abcd in self.n2a.items() :
            abcd = ", ".join("#" + t for t in abcd)
            if node in self.n2t :
                svg = svg.replace(' id="%s" ' % node,
                                  ' id="%s" data-abcd="%s" data-tree="#%s" '
                                  % (node, abcd, self.n2t[node]))
            else :
                svg = svg.replace(' id="%s" ' % node,
                                  ' id="%s" data-abcd="%s" ' % (node, abcd))
        return u"<div class='petrinet'>%s</div>" % svg

def build (abcd, node, net, gv, outfile, tpl=template_html, **args) :
    abcd = ABCD2HTML(node)
    pnet = Net2HTML(net, gv, abcd)
    d = {"filename" : node.st.filename,
         "css" : template_css,
         "jscode" : template_jscode,
         "headers" : template_headers,
         "abcd" : abcd.html(),
         "tree" : pnet.html(),
         "svg" : pnet.svg()}
    d.update(args)
    if tpl is not None and outfile :
        with codecs.open(outfile, "w", "utf-8") as out :
            out.write(tpl % d)
    return d



#!/usr/bin/env python

import subprocess
from distutils.core import setup

def doc_files () :
    import os, os.path
    result = {}
    for root, dirs, files in os.walk("doc") :
        target_dir = os.path.join("share/doc/python-snakes",
                                  *root.split(os.sep)[1:])
        for name in files :
            if target_dir not in result :
                result[target_dir] = []
            result[target_dir].append(os.path.join(root, name))
    return list(result.items())

def resources (root) :
    collected = ["*.txt", "*.html", "*.css", "*.js", "*.png", "*.jpg",
                 "*.ttf", "*.eot", "*.woff", "*.svg" ,"*.map"]
    import fnmatch, os.path
    result = []
    baselen = len(root.rstrip(os.sep).rsplit(os.sep, 1)[0]) + len(os.sep)
    def addfiles (_, dirname, filenames) :
        for name in filenames :
            for glob in collected :
                if fnmatch.fnmatch(name, glob) :
                    result.append(os.path.join(dirname[baselen:], name))
                    break
    os.path.walk(root, addfiles, None)
    return result

try :
    long_description=open("README").read()
except :
    long_description="""SNAKES is a general purpose Petri net Python
library allowing to define and execute most classes of Petri
nets. It features a plugin system to allow its extension. In
particular, plugins are provided to implement the operations
usually found in the PBC and M-nets family."""

if __name__ == "__main__" :
    emacs = [("share/emacs/site-lisp", ["utils/abcd-mode.el",
                                        "utils/abcd-mode.elc"])]
    setup(name="SNAKES",
          version=open("VERSION").read().strip(),
          description="SNAKES is the Net Algebra Kit for Editors and Simulators",
          long_description=long_description,
          author="Franck Pommereau",
          author_email="franck.pommereau@ibisc.univ-evry.fr",
          maintainer="Franck Pommereau",
          maintainer_email="franck.pommereau@ibisc.univ-evry.fr",
          url="http://http://www.ibisc.univ-evry.fr/~fpommereau/SNAKES/",
          scripts=["bin/abcd",
                   "bin/snkc",
                   "bin/snkd",
                   ],
          packages=["snakes",
                    "snakes.lang",
                    "snakes.lang.abcd",
                    "snakes.lang.ctlstar",
                    "snakes.lang.pylib",
                    "snakes.lang.python",
                    "snakes.plugins",
                    "snakes.utils",
                    "snakes.utils.abcd",
                    "snakes.utils.ctlstar",
                    "snakes.utils.simul",
                    ],
          package_data={"snakes.utils.abcd": resources("snakes/utils/abcd/resources/"),
                        "snakes.utils.simul": resources("snakes/utils/simul/resources/"),
                    },
          data_files=doc_files() + emacs,
          )
